# :coding: utf-8

import os

import pytest

import wiz.config
import wiz.definition
import wiz.graph
from wiz.utility import Requirement


@pytest.fixture(autouse=True)
def reset_configuration(mocker):
    """Ensure that no personal configuration is fetched during tests."""
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")

    # Reset configuration.
    wiz.config.fetch(refresh=True)


def test_scenario_1():
    """Compute packages for the following graph.

    Root
     |
     |--(A): A==0.2.0
     |   |
     |   `--(C >=0.3.2, <1): C==0.3.2
     |       |
     |       `--(D==0.1.0): D==0.1.0
     |
     `--(G): G==2.0.2
         |
         `--(B<0.2.0): B==0.1.0
             |
             |--(D>=0.1.0): D==0.1.4
             |   |
             |   `--(E>=2): E==2.3.0
             |       |
             |       `--(F>=0.2): F==1.0.0
             |
             `--(F>=1): F==1.0.0

    Expected: F==1.0.0, D==0.1.0, B==0.1.0, G==2.0.2, C==0.3.2, A==0.2.0

    """
    definition_mapping = {
        "A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "requirements": ["C>=0.3.2, <1"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0"
            })
        },
        "B": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.2.0"
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["D>=0.1.0", "F>=1"]
            })
        },
        "C": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.0.0"
            }),
            "0.3.2": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.3.2",
                "requirements": ["D==0.1.0"]
            })
        },
        "D": {
            "0.1.4": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.4",
                "requirements": ["E>=2"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0"
            })
        },
        "E": {
            "2.3.0": wiz.definition.Definition({
                "identifier": "E",
                "version": "2.3.0",
                "requirements": ["F>=0.2"]
            })
        },
        "F": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "1.0.0"
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "0.9.0"
            })
        },
        "G": {
            "2.0.2": wiz.definition.Definition({
                "identifier": "G",
                "version": "2.0.2",
                "requirements": ["B<0.2.0"]
            }),
            "2.0.1": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.1"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([Requirement("A"), Requirement("G")])

    assert len(packages) == 6
    assert packages[0].identifier == "F==1.0.0"
    assert packages[1].identifier == "D==0.1.0"
    assert packages[2].identifier == "B==0.1.0"
    assert packages[3].identifier == "G==2.0.2"
    assert packages[4].identifier == "C==0.3.2"
    assert packages[5].identifier == "A==0.2.0"


def test_scenario_2():
    """Fail to compute packages for the following graph.

    Root
     |
     |--(A): A==0.2.0
     |   |
     |   `--(C >=0.3.2, <1): C==0.3.2
     |       |
     |       `--(D==0.1.0): D==0.1.0
     |
     `--(G): G==2.0.2
         |
         `--(B<0.2.0): B==0.1.0
             |
             |--(D>0.1.0): D==0.1.4
             |   |
             |   `--(E>=2): E==2.3.0
             |       |
             |       `--(F>=0.2): F==1.0.0
             |
             `--(F>=1): F==1.0.0

    Expected: Unable to compute due to requirement compatibility between
    'D>0.1.0' and 'D==0.1.0'.

    """
    definition_mapping = {
        "A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "requirements": ["C>=0.3.2, <1"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0"
            })
        },
        "B": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.2.0"
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["D>0.1.0", "F>=1"]
            })
        },
        "C": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.0.0"
            }),
            "0.3.2": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.3.2",
                "requirements": ["D==0.1.0"]
            })
        },
        "D": {
            "0.1.4": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.4",
                "requirements": ["E>=2"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0"
            })
        },
        "E": {
            "2.3.0": wiz.definition.Definition({
                "identifier": "E",
                "version": "2.3.0",
                "requirements": ["F>=0.2"]
            })
        },
        "F": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "1.0.0"
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "0.9.0"
            })
        },
        "G": {
            "2.0.2": wiz.definition.Definition({
                "identifier": "G",
                "version": "2.0.2",
                "requirements": ["B<0.2.0"]
            }),
            "2.0.1": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.1"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A"), Requirement("G")])

    assert (
        "The dependency graph could not be resolved due to the "
        "following requirement conflicts:\n"
        "  * ::D ==0.1.0 \t[C==0.3.2]\n"
        "  * ::D >0.1.0 \t[B==0.1.0]"
    ) in str(error.value)


def test_scenario_3():
    """Compute packages for the following graph.

    In a situation with several solutions, the solution which guaranty the
    conservation of the node nearest to the top level is chosen

    Root
     |
     |--(A): A==1.0.0
     |   |
     |   `--(B<1): B==0.9.0
     |
     `--(B): B==1.0.0
         |
         `--(A<1): A==0.9.0

    Expected: B==0.9.0, A==1.0.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "requirements": ["B<1"]
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.9.0"
            })
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0",
                "requirements": ["A<1"]
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.9.0"
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([Requirement("A"), Requirement("B")])

    assert len(packages) == 2
    assert packages[0].identifier == "B==0.9.0"
    assert packages[1].identifier == "A==1.0.0"


def test_scenario_4():
    """Compute packages for the following graph.

    When only the identifier of a definition with several variants is required,
    all variants are added to the graph which is then divided into as many
    graphs as there are variants. The graph are ordered following the order of
    the variants and the first graph to resolve is the solution.

    Root
     |
     |--(A): A[V1]==1.0.0
     |   |
     |   `--(B >=1, <2): B==1.0.0
     |
     |--(A): A[V2]==1.0.0
     |   |
     |   `--(B >=2, <3): B==2.0.0
     |
     |--(A): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     `--(A): A[V4]==1.0.0
         |
         `--(B >=4, <5): B==4.0.0

    Expected: B==4.0.0, A[V4]==1.0.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V4",
                        "requirements": ["B >=4, <5"]
                    },
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            }),
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
            "4.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "4.0.0"
            }),
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([Requirement("A")])

    assert len(packages) == 2
    assert packages[0].identifier == "B==4.0.0"
    assert packages[1].identifier == "A[V4]==1.0.0"


def test_scenario_5():
    """Compute packages for the following graph.

    When a specific variant of a definition is required, only one graph with
    this exact variant is added to the graph.

    Root
     |
     `--(A[V1]): A[V1]==1.0.0
         |
         `--(B >=1, <2): B==1.0.0

    Expected: B==1.0.0, A[V1]==1.0.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V4",
                        "requirements": ["B >=4, <5"]
                    },
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            }),
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
            "4.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "4.0.0"
            }),
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([Requirement("A[V1]")])

    assert len(packages) == 2
    assert packages[0].identifier == "B==1.0.0"
    assert packages[1].identifier == "A[V1]==1.0.0"


def test_scenario_6():
    """Compute packages for the following graph.

    Like the scenario 4, we end up with as many graph as there are variants. But
    the additional requirement makes the 2 first graphs fail so that only the
    3rd graph is resolved.

    Root
     |
     |--(A): A[V1]==1.0.0
     |   |
     |   `--(B >=1, <2): B==1.0.0
     |
     |--(A): A[V2]==1.0.0
     |   |
     |   `--(B >=2, <3): B==2.0.0
     |
     |--(A): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     |--(A): A[V4]==1.0.0
     |   |
     |   `--(B >=4, <5): B==4.0.0
     |
     `--(B==2.*): B==2.0.0

    Expected: B==2.0.0, A[V2]==1.0.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V4",
                        "requirements": ["B >=4, <5"]
                    },
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            }),
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
            "4.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "4.0.0"
            }),
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B==2.*")
    ])

    assert len(packages) == 2
    assert packages[0].identifier == "B==2.0.0"
    assert packages[1].identifier == "A[V2]==1.0.0"


def test_scenario_7():
    """Compute packages for the following graph.

    The combined requirement of packages can lead to the addition of a different
    package version to the graph during the conflict resolution process.

    Root
     |
     |--(A<=0.3.0): A==0.3.0
     |
     |--(B): B==0.1.0
     |   |
     |   `--(A !=0.3.0): A==1.0.0
     |
     `--(C): C==0.1.0
         |
         `--(A <1): A==0.9.0

    Expected: C==0.1.0, B==0.1.0, A==0.2.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0"
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.9.0"
            }),
            "0.3.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.3.0"
            }),
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0"
            })
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["A !=0.3.0"]
            })
        },
        "C": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.1.0",
                "requirements": ["A <1"]
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A<=0.3.0"), Requirement("B"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C==0.1.0"
    assert packages[1].identifier == "B==0.1.0"
    assert packages[2].identifier == "A==0.2.0"


def test_scenario_8():
    """Compute packages for the following graph.

    When conflicts parents are themselves conflicting, they should be discarded
    so that the next conflict is taken care of first.

    Root
     |
     |--(A): A==1.0.0
     |   |
     |   `--(C>=1): C== 1.0.0
     |
     `--(B): B==0.1.0
         |
         `--(A <1): A== 0.9.0
             |
             `--(C <1): C== 0.9.0

    Expected: A==0.9.0, C==0.9.0, B==0.1.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "requirements": ["C>=1"]
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.9.0",
                "requirements": ["C<1"]
            })
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["A<1"],

            })
        },
        "C": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.0.0"
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.9.0"
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C==0.9.0"
    assert packages[1].identifier == "B==0.1.0"
    assert packages[2].identifier == "A==0.9.0"


def test_scenario_9():
    """Compute packages for the following graph.

    Like the scenario 8 with different order in the graph.

    Root
     |
     |--(A <1): A== 0.9.0
     |   |
     |   `--(C <1): C== 0.9.0
     |
     `--(B): B==0.1.0
         |
         `--(A): A== 1.0.0
             |
             `--(C>=1): C== 1.0.0

    Expected: B==0.1.0, C==0.9.0, A==0.9.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "requirements": ["C>=1"]
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.9.0",
                "requirements": ["C<1"]
            })
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["A"],

            })
        },
        "C": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.0.0"
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.9.0"
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A <1"), Requirement("B")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C==0.9.0"
    assert packages[1].identifier == "B==0.1.0"
    assert packages[2].identifier == "A==0.9.0"


def test_scenario_10():
    """Compute packages for the following graph.

    Like the scenario 7, a new node is added to the graph during the conflict
    resolution process, but this time the new node leads to a division of the
    graph.

    Root
     |
     |--(A<=0.3.0): A==0.3.0
     |
     `--(B): B==0.1.0
         |
         `--(A !=0.3.0): A==1.0.0

    Expected: C[V3]==1.0.0, B==0.1.0, A==0.2.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0"
            }),
            "0.3.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.3.0"
            }),
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "requirements": ["C"]
            })
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["A !=0.3.0"]
            })
        },
        "C": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                    },
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            }),
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A<=0.3.0"), Requirement("B")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C[V3]==1.0.0"
    assert packages[1].identifier == "B==0.1.0"
    assert packages[2].identifier == "A==0.2.0"


def test_scenario_11():
    """Compute packages for the following graph.

    When a definition variant is present more than other variants from the same
    definition in the graph, it has priority.

    Root
     |
     |--(A): A[V1]==1.0.0
     |   |
     |   `--(B >=1, <2): B==1.0.0
     |
     |--(A): A[V2]==1.0.0
     |   |
     |   `--(B >=2, <3): B==2.0.0
     |
     |--(A): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     `--(C): C
         |
         `--(A[V2]): A[V2]==1.0.0
             |
             `--(B >=2, <3): B==2.0.0

    Expected: C, B==2.0.0, A[V2]==1.0.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            })
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "requirements": ["A[V2]"]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C"
    assert packages[1].identifier == "B==2.0.0"
    assert packages[2].identifier == "A[V2]==1.0.0"


def test_scenario_12():
    """Compute packages for the following graph.

    When two versions of a definition are added to the graph with all their
    respective variants, the conflict is resolved for the variant with the
    highest priority.

    Root
     |
     |--(A): A[V1]==1.0.0
     |   |
     |   `--(B >=1, <2): B==1.0.0
     |
     |--(A): A[V2]==1.0.0
     |   |
     |   `--(B >=2, <3): B==2.0.0
     |
     |--(A): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     `--(C): C
         |
         |--(A==0.5.0): A[V1]==0.5.0
         |   |
         |   `--(B >=1, <2): B==1.0.0
         |
         |--(A==0.5.0): A[V2]==0.5.0
         |   |
         |   `--(B >=1, <2): B==1.0.0
         |
         `--(A==0.5.0): A[V3]==0.5.0
             |
             `--(B >=1, <2): B==1.0.0

    Expected: C, B==3.0.0, A[V3]==0.5.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            }),
            "0.5.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.5.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            }),
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "requirements": ["A==0.5.0"]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C"
    assert packages[1].identifier == "B==3.0.0"
    assert packages[2].identifier == "A[V3]==0.5.0"


def test_scenario_13():
    """Compute packages for the following graph.

    When computing a graph combination, all variant nodes that should not be
    included are removed from the graph. During the removal process, parents
    are relinked to ensure that all requirements are still respected. If a
    parent cannot be linked to any existing nodes in the graph, the combination
    cannot be resolved.

    In this example, "A[V3]==0.1.0" is being computed first so all other nodes
    belonging to the definition identifier "A" are removed, which makes it
    impossible for "C"'s requirement to be fulfilled.

    Root
     |
     |--(A): A[V1]==1.0.0
     |   |
     |   `--(B >=1, <2): B==1.0.0
     |
     |--(A): A[V2]==1.0.0
     |   |
     |   `--(B >=2, <3): B==2.0.0
     |
     |--(A): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     `--(C): C
         |
         `--(A==0.5.0): A[V1]==0.5.0
             |
             `--(B >=1, <2): B==1.0.0

    Expected: C, B==1.0.0, A[V1]==0.5.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            }),
            "0.5.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.5.0",
                "variants": [
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            }),
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "requirements": ["A==0.5.0"]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C"
    assert packages[1].identifier == "B==1.0.0"
    assert packages[2].identifier == "A[V1]==0.5.0"


def test_scenario_14():
    """Compute packages for the following graph.

    When several packages with variants are added to the graph, the variant
    with the highest priority is returned for each package if no conflict
    appear.

    Root
     |
     |--(A): A[V1]
     |
     |--(A): A[V2]
     |
     |--(A): A[V3]
     |
     |--(B): B[V1]
     |
     |--(B): B[V2]
     |
     |--(B): B[V3]
     |
     |--(B): B[V4]
     |
     |--(C): C[V1]
     |
     `--(C): C[V2]

    Expected: C[V2], B[V4], A[V3]

    """
    definition_mapping = {
        "A": {
            "-": wiz.definition.Definition({
                "identifier": "A",
                "variants": [
                    {
                        "identifier": "V3",
                    },
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "B": {
            "-": wiz.definition.Definition({
                "identifier": "B",
                "variants": [
                    {
                        "identifier": "V4",
                    },
                    {
                        "identifier": "V3",
                    },
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "variants": [
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C[V2]"
    assert packages[1].identifier == "B[V4]"
    assert packages[2].identifier == "A[V3]"


def test_scenario_15():
    """Compute packages for the following graph.

    If a definition variant contains an error, the combinations containing the
    error are discarded until a variant of the definition without errors is
    picked.

    Root
     |
     |--(A): A[V1]
     |
     |--(A): A[V2]
     |
     |--(A): A[V3]
     |
     |--(B): B[V1]
     |
     |--(B): B[V2]
     |
     |--(B): B[V3]
     |
     |--(B): B[V4]
     |   |
     |   `--(incorrect): ERROR!
     |
     |--(C): C[V1]
     |
     `--(C): C[V2]

    Expected: C[V2], B[V3], A[V3]

    """
    definition_mapping = {
        "A": {
            "-": wiz.definition.Definition({
                "identifier": "A",
                "variants": [
                    {
                        "identifier": "V3",
                    },
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "B": {
            "-": wiz.definition.Definition({
                "identifier": "B",
                "variants": [
                    {
                        "identifier": "V4",
                        "requirements": ["incorrect"]
                    },
                    {
                        "identifier": "V3",
                    },
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "variants": [
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C[V2]"
    assert packages[1].identifier == "B[V3]"
    assert packages[2].identifier == "A[V3]"


def test_scenario_16():
    """Compute packages for the following graph.

    For the same example as scenario 15, if the package 'B[V4]' has a
    conflicting requirement instead of an error, the combinations containing the
    conflict are discarded until a variant of the definition without conflicts
    is picked.

    Root
     |
     |--(A): A[V1]
     |
     |--(A): A[V2]
     |
     |--(A): A[V3]
     |
     |--(B): B[V1]
     |
     |--(B): B[V2]
     |
     |--(B): B[V3]
     |
     |--(B): B[V4]
     |   |
     |   `--(D>=1): D==1.0.0
     |
     |--(C): C[V1]
     |
     |--(C): C[V2]
     |
     `--(D<1): D==0.1.0

    Expected: C[V2], B[V3], A[V3], D==0.1.0

    """
    definition_mapping = {
        "A": {
            "-": wiz.definition.Definition({
                "identifier": "A",
                "variants": [
                    {
                        "identifier": "V3",
                    },
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "B": {
            "-": wiz.definition.Definition({
                "identifier": "B",
                "variants": [
                    {
                        "identifier": "V4",
                        "requirements": ["D>=1"]
                    },
                    {
                        "identifier": "V3",
                    },
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "variants": [
                    {
                        "identifier": "V2",
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "D": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "D",
                "version": "1.0.0"
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B"), Requirement("C"), Requirement("D<1")
    ])

    assert len(packages) == 4
    assert packages[0].identifier == "D==0.1.0"
    assert packages[1].identifier == "C[V2]"
    assert packages[2].identifier == "B[V3]"
    assert packages[3].identifier == "A[V3]"


def test_scenario_17():
    """Compute packages for the following graph.

    When a package has a condition which is not fulfilled, it is not included
    in the resolved packages.

    Root
     |
     |--(A): A==1.1.0
     |
     `--(B): B==1.0.0 (Condition: A < 1)
         |
         `--(C > 0.1.0): C==0.5.0

    Expected: A==1.1.0

    """
    definition_mapping = {
        "A": {
            "1.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.1.0",
            }),
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
            }),
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0",
                "conditions": ["A < 1"],
                "requirements": ["C > 0.1.0"],
            })
        },
        "C": {
            "0.5.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.5.0"
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B")
    ])

    assert len(packages) == 1
    assert packages[0].identifier == "A==1.1.0"


def test_scenario_18():
    """Compute packages for the following graph.

    When a package has a condition which is fulfilled, it is properly included
    in the resolved packages with expected distance priority.

    Root
     |
     |--(A): A==1.1.0
     |
     `--(B): B==1.0.0 (Condition: A > 1)
         |
         `--(C > 0.1.0): C==0.5.0

    Expected: A==1.1.0, B==1.0.0, C==0.5.0

    """
    definition_mapping = {
        "A": {
            "1.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.1.0",
            }),
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
            }),
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0",
                "conditions": ["A > 1"],
                "requirements": ["C > 0.1.0"],

            })
        },
        "C": {
            "0.5.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.5.0"
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C==0.5.0"
    assert packages[1].identifier == "B==1.0.0"
    assert packages[2].identifier == "A==1.1.0"


def test_scenario_19():
    """Compute packages for the following graph.

    A package condition can be fulfilled by packages deep in the graph.

    Root
     |
     |--(A): A==0.2.0 (Condition: E)
     |   |
     |   `--(C >=0.3.2, <1): C==0.3.2
     |
     `--(G): G==2.0.2
         |
         `--(B<0.2.0): B==0.1.0
             |
             `--(D>=0.1.0): D==0.1.4
                 |
                 `--(E>=2): E==2.3.0

    Expected: E==2.3.0, D==0.1.4, B==0.1.0, G==2.0.2, C==0.3.2, A==0.2.0

    """
    definition_mapping = {
        "A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "conditions": ["E"],
                "requirements": ["C >=0.3.2, <1"]
            }),
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["D >=0.1.0"],
            })
        },
        "C": {
            "0.3.2": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.3.2"
            })
        },
        "D": {
            "0.1.4": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.4",
                "requirements": ["E >=2"]
            })
        },
        "E": {
            "2.3.0": wiz.definition.Definition({
                "identifier": "E",
                "version": "2.3.0"
            })
        },
        "G": {
            "2.0.2": wiz.definition.Definition({
                "identifier": "G",
                "version": "2.0.2",
                "requirements": ["B <0.2.0"]
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("G")
    ])

    assert len(packages) == 6
    assert packages[0].identifier == "E==2.3.0"
    assert packages[1].identifier == "D==0.1.4"
    assert packages[2].identifier == "B==0.1.0"
    assert packages[3].identifier == "G==2.0.2"
    assert packages[4].identifier == "C==0.3.2"
    assert packages[5].identifier == "A==0.2.0"


def test_scenario_20():
    """Compute packages for the following graph.

    For the same example as the scenario 19, if the condition can not be
    fulfilled, the package is not included in the resolved packages.

    Root
     |
     |--(A): A==0.2.0 (Condition: E, F)
     |   |
     |   `--(C >=0.3.2, <1): C==0.3.2
     |
     `--(G): G==2.0.2
         |
         `--(B<0.2.0): B==0.1.0
             |
             `--(D>=0.1.0): D==0.1.4
                 |
                 `--(E>=2): E==2.3.0

    Expected: E==2.3.0, D==0.1.4, B==0.1.0, G==2.0.2

    """
    definition_mapping = {
        "A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "conditions": ["E", "F"],
                "requirements": ["C >=0.3.2, <1"]
            }),
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["D >=0.1.0"],
            })
        },
        "C": {
            "0.3.2": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.3.2"
            })
        },
        "D": {
            "0.1.4": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.4",
                "requirements": ["E >=2"]
            })
        },
        "E": {
            "2.3.0": wiz.definition.Definition({
                "identifier": "E",
                "version": "2.3.0"
            }),
        },
        "F": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "0.1.0"
            }),
        },
        "G": {
            "2.0.2": wiz.definition.Definition({
                "identifier": "G",
                "version": "2.0.2",
                "requirements": ["B <0.2.0"]
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("G")
    ])

    assert len(packages) == 4
    assert packages[0].identifier == "E==2.3.0"
    assert packages[1].identifier == "D==0.1.4"
    assert packages[2].identifier == "B==0.1.0"
    assert packages[3].identifier == "G==2.0.2"


def test_scenario_21():
    """Compute packages for the following graph.

    A package which has been added to the graph when its conditions were
    fulfilled will be removed once conflict resolution has resulted in the
    removal of a node the package is conditioned by.

    Root
     |
     |--(A): A==0.2.0 (Condition: E)
     |   |
     |   `--(C >=0.3.2, <1): C==0.3.2
     |       |
     |       `--(D==0.1.0): D==0.1.0
     |
     `--(G): G==2.0.2
         |
         `--(B<0.2.0): B==0.1.0
             |
             |--(D>=0.1.0): D==0.1.4
             |   |
             |   `--(E>=2): E==2.3.0
             |       |
             |       `--(F>=0.2): F==1.0.0
             |
             `--(F>=1): F==1.0.0

    Expected: F==1.0.0, D==0.1.0, B==0.1.0, G==2.0.2

    """
    definition_mapping = {
        "A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "requirements": ["C>=0.3.2, <1"],
                "conditions": ["E"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0"
            })
        },
        "B": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.2.0"
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["D>=0.1.0", "F>=1"]
            })
        },
        "C": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.0.0"
            }),
            "0.3.2": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.3.2",
                "requirements": ["D==0.1.0"]
            })
        },
        "D": {
            "0.1.4": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.4",
                "requirements": ["E>=2"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0"
            })
        },
        "E": {
            "2.3.0": wiz.definition.Definition({
                "identifier": "E",
                "version": "2.3.0",
                "requirements": ["F>=0.2"]
            })
        },
        "F": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "1.0.0"
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "0.9.0"
            })
        },
        "G": {
            "2.0.2": wiz.definition.Definition({
                "identifier": "G",
                "version": "2.0.2",
                "requirements": ["B<0.2.0"]
            }),
            "2.0.1": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.1"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([Requirement("A"), Requirement("G")])

    assert len(packages) == 4
    assert packages[0].identifier == "F==1.0.0"
    assert packages[1].identifier == "D==0.1.0"
    assert packages[2].identifier == "B==0.1.0"
    assert packages[3].identifier == "G==2.0.2"


def test_scenario_22():
    """Fail to compute packages for the following graph.

    Root
     |
     `--(A): Namespace1::A==0.1.0 || Namespace2::A==0.2.0

    Expected: Unable to guess default namespace for 'A'

    """
    definition_mapping = {
        "__namespace__": {
            "A": ["Namespace1", "Namespace2"]
        },
        "Namespace1::A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "namespace": "Namespace1"
            })
        },
        "Namespace2::A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "namespace": "Namespace2"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A")])

    assert (
        "The dependency graph could not be resolved due to the following "
        "error(s):\n"
        "  * root: Cannot guess default namespace for 'A' [available: "
        "Namespace1, Namespace2]."
    ) in str(error.value)


def test_scenario_23():
    """Compute packages for the following graph.

    Initial namespace counter can be used to give hint when looking
    for default namespace belonging to the other definitions.

    Root
     |
     |--(A): Namespace1::A
     |
     |--(B): Namespace1::B==0.1.0 || Namespace2::B==0.2.0
     |
     |--(C): Namespace3::C==0.1.0 || Namespace4::C==0.2.0
     |
     `--(D): Namespace4::D

    Expected:
    Namespace4::D, Namespace4::C==0.2.0, Namespace1::B==0.1.0,  Namespace1::A

    """
    definition_mapping = {
        "__namespace__": {
            "A": ["Namespace1"],
            "B": ["Namespace1", "Namespace2"],
            "C": ["Namespace3", "Namespace4"],
            "D": ["Namespace4"],
        },
        "Namespace1::A": {
            "-": wiz.definition.Definition({
                "identifier": "A",
                "namespace": "Namespace1"
            })
        },
        "Namespace1::B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "namespace": "Namespace1"
            })
        },
        "Namespace2::B": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.2.0",
                "namespace": "Namespace2"
            })
        },
        "Namespace3::C": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.1.0",
                "namespace": "Namespace3"
            })
        },
        "Namespace4::C": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.2.0",
                "namespace": "Namespace4"
            })
        },
        "Namespace4::D": {
            "-": wiz.definition.Definition({
                "identifier": "D",
                "namespace": "Namespace4"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    requirements = [
        Requirement("A"), Requirement("B"), Requirement("C"), Requirement("D")
    ]
    namespace_counter = wiz.utility.compute_namespace_counter(
        requirements, definition_mapping
    )

    with pytest.raises(wiz.exception.GraphInvalidNodesError) as error:
        resolver.compute_packages(requirements)

    assert (
        "The dependency graph could not be resolved due to the following "
        "error(s):\n"
        "  * root: Cannot guess default namespace for 'C' [available: "
        "Namespace3, Namespace4]."
    ) in str(error.value)

    packages = resolver.compute_packages(
        requirements, namespace_counter=namespace_counter
    )

    assert len(packages) == 4
    assert packages[0].identifier == "Namespace4::D"
    assert packages[1].identifier == "Namespace4::C==0.2.0"
    assert packages[2].identifier == "Namespace1::B==0.1.0"
    assert packages[3].identifier == "Namespace1::A"


def test_scenario_24():
    """Compute packages for the following graph.

    When a definition exists with and without a namespace, the one without a
    namespace is selected automatically by default.

    Root
     |
     `--(A): A==0.1.0 || Namespace1::A==0.2.0

    Expected: A==0.1.0

    """
    definition_mapping = {
        "__namespace__": {
            "A": ["Namespace1"]
        },
        "A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0"
            })
        },
        "Namespace1::A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "namespace": "Namespace1"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    packages = resolver.compute_packages([Requirement("A")])

    assert len(packages) == 1
    assert packages[0].identifier == "A==0.1.0"


def test_scenario_25():
    """Compute packages for the following graph.

    When a definition exists with and without a namespace, other definitions
    using the same namespace within the request will lead to the selection of
    the one with the namespace if an initial counter is being used.

    Root
     |
     |--(A): A==0.1.0 || Namespace1::A==0.2.0
     |
     `--(B): Namespace1::B==0.1.0

    Expected: Namespace1::B==0.1.0, Namespace1::A==0.1.0

    """
    definition_mapping = {
        "__namespace__": {
            "A": ["Namespace1"],
            "B": ["Namespace1"]
        },
        "A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0"
            })
        },
        "Namespace1::A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "namespace": "Namespace1"
            })
        },
        "Namespace1::B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "namespace": "Namespace1"
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    requirements = [Requirement("A"), Requirement("B")]
    namespace_counter = wiz.utility.compute_namespace_counter(
        requirements, definition_mapping
    )

    packages = resolver.compute_packages(requirements)

    assert len(packages) == 2
    assert packages[0].identifier == "Namespace1::B==0.1.0"
    assert packages[1].identifier == "A==0.1.0"

    packages = resolver.compute_packages(
        requirements, namespace_counter=namespace_counter
    )

    assert len(packages) == 2
    assert packages[0].identifier == "Namespace1::B==0.1.0"
    assert packages[1].identifier == "Namespace1::A==0.2.0"


def test_scenario_26():
    """Compute packages for the following graph.

    When a definition exists with and without a namespace, the one without a
    namespace can be explicitly called.

    Root
     |
     |--(::A): A==0.1.0 || Namespace1::A==0.2.0
     |
     `--(B): Namespace1::B==0.1.0

    Expected: Namespace1::B==0.1.0, A==0.1.0

    """
    definition_mapping = {
        "__namespace__": {
            "A": ["Namespace1"],
            "B": ["Namespace1"]
        },
        "A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0"
            })
        },
        "Namespace1::A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "namespace": "Namespace1"
            })
        },
        "Namespace1::B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "namespace": "Namespace1"
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    packages = resolver.compute_packages([Requirement("::A"), Requirement("B")])

    assert len(packages) == 2
    assert packages[0].identifier == "Namespace1::B==0.1.0"
    assert packages[1].identifier == "A==0.1.0"


def test_scenario_27():
    """Compute packages for the following graph.

    A package which has been added to the graph when its conditions were
    fulfilled will be kept when conflict resolution has resulted in the
    removal of a node the package is conditioned as long as another node which
    fulfill the same condition remains in the graph.

    Root
     |
     |--(A): A==0.1.0 (Condition: B)
     |   |
     |   `--(B >= 0.1.0, < 1) B==0.1.0
     |
     `--(B): B==1.0.0

    Expected: B==0.1.0, A==0.1.0

    """
    definition_mapping = {
        "A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "conditions": ["B"],
                "requirements": ["B >= 0.1.0, < 1"],
            })
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    packages = resolver.compute_packages([Requirement("A"), Requirement("B")])

    assert len(packages) == 2
    assert packages[0].identifier == "B==0.1.0"
    assert packages[1].identifier == "A==0.1.0"


def test_scenario_28():
    """Compute packages for the following graph.

    Like for the scenario 22, the package "A" has several namespaces available.
    But as one of the namespace is identical to the definition identifier, it
    will be selected by default.

    Root
     |
     `--(A): Namespace1::A==0.1.0 || A::A==0.2.0

    Expected: A::A==0.2.0

    """
    definition_mapping = {
        "__namespace__": {
            "A": ["Namespace1", "A"]
        },
        "Namespace1::A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "namespace": "Namespace1"
            })
        },
        "A::A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "namespace": "A"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    packages = resolver.compute_packages([Requirement("A")])

    assert len(packages) == 1
    assert packages[0].identifier == "A::A==0.2.0"


def test_scenario_29():
    """Compute packages for the following graph.

    Like for the scenario 7, the combined requirement of packages can lead to
    the addition of a different package version to the graph during the conflict
    resolution process. But when it happens with a definition with variant, all
    variants must be added.

    Root
     |
     |--(B): B==0.1.0
     |   |
     |   |--(A !=0.3.0): A[V1]==1.0.0
     |   |
     |   `--(A !=0.3.0): A[V2]==1.0.0
     |
     `--(C): C==0.1.0
         |
         |--(A <=0.3.0): A[V3]==0.3.0
         |
         |--(A <=0.3.0): A[V2]==0.3.0
         |
         `--(A <=0.3.0): A[V1]==0.3.0

    Expected: C==0.1.0, A[V1]==0.2.0, B==0.1.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V2"
                    },
                    {
                        "identifier": "V1"
                    }
                ]
            }),
            "0.3.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.3.0",
                "variants": [
                    {
                        "identifier": "V3"
                    },
                    {
                        "identifier": "V2"
                    },
                    {
                        "identifier": "V1"
                    }
                ]
            }),
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "variants": [
                    {
                        "identifier": "V2",
                        "requirements": ["incorrect"]
                    },
                    {
                        "identifier": "V1",
                    }
                ]
            })
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["A !=0.3.0"]
            })
        },
        "C": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.1.0",
                "requirements": ["A <=0.3.0"]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("B"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C==0.1.0"
    assert packages[1].identifier == "A[V1]==0.2.0"
    assert packages[2].identifier == "B==0.1.0"


def test_scenario_30():
    """Fail to compute packages for the following graph.

    If a definition contains an error, it will lead to the failure of the graph
    when no other possible combinations are available.

    Root
     |
     |--(A): A==0.1.0
     |   |
     |   `--(incorrect): ERROR!
     |
     `--(B): B==0.1.0

    Expected: The requirement 'incorrect' could not be resolved.

    """
    definition_mapping = {
        "A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "requirements": ["incorrect"]
            }),
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A"), Requirement("B")])

    assert (
        "The dependency graph could not be resolved due to the following "
        "error(s):\n"
        "  * A==0.1.0: The requirement 'incorrect' could not be resolved."
    ) in str(error.value)


def test_scenario_31():
    """Fail to compute packages for the following graph.

    Like scenario 15, when a definition variant contains an error, the next
    combination will try to detect and raise when this erroneous node is
    encountered again. However, when there are no combination that can be
    resolved, it will still fail.

    Root
     |
     |--(A): A[V1]==0.1.0
     |   |
     |   `--(C): C==0.1.0
     |       |
     |       `--(incorrect): ERROR!
     |
     |--(A): A[V2]==0.1.0
     |   |
     |   `--(C): C==0.1.0
     |       |
     |       `--(incorrect): ERROR!
     |
     `--(B): B==0.1.0

    Expected: The requirement 'incorrect' could not be resolved.

    """
    definition_mapping = {
        "A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "variants": [
                    {
                        "identifier": "V2",
                        "requirements": ["C"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["C"]

                    },
                ]
            }),
        },
        "B": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
            })
        },
        "C": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.1.0",
                "requirements": ["incorrect"]
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A"), Requirement("B")])

    assert (
        "The dependency graph could not be resolved due to the following "
        "error(s):\n"
        "  * C==0.1.0: The requirement 'incorrect' could not be resolved."
    ) in str(error.value)


def test_scenario_32():
    """Fail to compute packages for the following graph.

    When a conflict is detected, it is being stored to prevent resolving a
    combination graph which involve the same conflict. However, when there are
    no combination that can be resolved, it will still fail.

    Root
     |
     |--(A): A[V1]==0.1.0
     |   |
     |   `--(C): C==0.1.0
     |       |
     |       `--(B<1): B==0.1.0
     |
     |--(A): A[V2]==0.1.0
     |   |
     |   `--(C): C==0.1.0
     |       |
     |       `--(B<1): B==0.1.0
     |
     `--(B>1): B==1.1.0

    Expected: Unable to compute due to requirement compatibility between
    'B<1' and 'B>1'.

    """
    definition_mapping = {
        "A": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "variants": [
                    {
                        "identifier": "V2",
                        "requirements": ["C"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["C"]

                    },
                ]
            }),
        },
        "B": {
            "1.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.1.0",
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
            })
        },
        "C": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.1.0",
                "requirements": ["B<1"]
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A"), Requirement("B>1")])

    assert (
        "The dependency graph could not be resolved due to the "
        "following requirement conflicts:\n"
        "  * ::B <1 \t[C==0.1.0]\n"
        "  * ::B >1 \t[root]"
    ) in str(error.value)


def test_scenario_33():
    """Compute packages for the following graph.

    Like scenario 2, the graph resolution will fail with the all node versions
    fetched in the graph. However, it can drop down the version of package
    with conflicting requirement to a version which does not conflict.

    Root
     |
     |--(A): A==0.2.0
     |   |
     |   `--(C >=0.3.0, <1): C==0.3.2
     |       |
     |       `--(D==0.1.0): D==0.1.0
     |
     `--(G): G==2.0.2
         |
         `--(B<0.2.0): B==0.1.0
             |
             |--(D>0.1.0): D==0.1.4
             |   |
             |   `--(E>=2): E==2.3.0
             |       |
             |       `--(F>=0.2): F==1.0.0
             |
             `--(F>=1): F==1.0.0

    Expected: F==1.0.0, E==2.3.0, D==0.1.4, B==0.1.0, G==2.0.2, C==0.3.0
    A==0.2.0

    """
    definition_mapping = {
        "A": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "requirements": ["C>=0.3.0, <1"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0"
            })
        },
        "B": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.2.0"
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["D>0.1.0", "F>=1"]
            })
        },
        "C": {
            "0.3.2": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.3.2",
                "requirements": ["D==0.1.0"]
            }),
            "0.3.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.3.0"
            })
        },
        "D": {
            "0.1.4": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.4",
                "requirements": ["E>=2"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0"
            })
        },
        "E": {
            "2.3.0": wiz.definition.Definition({
                "identifier": "E",
                "version": "2.3.0",
                "requirements": ["F>=0.2"]
            })
        },
        "F": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "1.0.0"
            }),
            "0.9.0": wiz.definition.Definition({
                "identifier": "F",
                "version": "0.9.0"
            })
        },
        "G": {
            "2.0.2": wiz.definition.Definition({
                "identifier": "G",
                "version": "2.0.2",
                "requirements": ["B<0.2.0"]
            }),
            "2.0.1": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.1"
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([Requirement("A"), Requirement("G")])

    assert len(packages) == 7
    assert packages[0].identifier == "F==1.0.0"
    assert packages[1].identifier == "E==2.3.0"
    assert packages[2].identifier == "D==0.1.4"
    assert packages[3].identifier == "B==0.1.0"
    assert packages[4].identifier == "G==2.0.2"
    assert packages[5].identifier == "C==0.3.0"
    assert packages[6].identifier == "A==0.2.0"


def test_scenario_34():
    """Fail to compute packages for the following graph.

    Like scenario 13, parents are relinked when computing a graph combination
    to ensure that all requirements are still respected.

    But in this example, two requirements are explicitly requesting conflicting
    variants: "A[V2]" and "A[V3]".

    Root
     |
     |--(A[V3]): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     `--(C): C
         |
         `--(A[V2]): A[V2]==1.0.0
             |
             `--(B >=2, <3): B==2.0.0

    Expected: Unable to compute due to requirement compatibility between
    'A[V2]' and 'A[V3]'.

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    }
                ]
            })
        },
        "B": {
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "requirements": ["A[V2]"]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A[V3]"), Requirement("C")])

    assert (
        "The dependency graph could not be resolved due to the following "
        "requirement conflicts:\n"
        "  * ::A[V2] \t[C]\n"
        "  * ::A[V3] \t[root]"
    ) in str(error.value)


def test_scenario_35():
    """Fail to compute packages for the following graph.

    Like scenario 34, two requirements are explicitly requesting conflicting
    variants: "A[V2]" and "A[V3]".

    Root
     |
     |--(A): A[V1]==1.0.0
     |   |
     |   `--(B >=1, <2): B==1.0.0
     |
     |--(A): A[V2]==1.0.0
     |   |
     |   `--(B >=2, <3): B==2.0.0
     |
     |--(A): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     |--(C): C
     |   |
     |   `--(A[V2]): A[V2]==1.0.0
     |       |
     |       `--(B >=1, <2): B==1.0.0
     |
     `--(D): D==0.1.0
         |
         `--(A[V3]): A[V3]==1.0.0

    Expected: Unable to compute due to requirement compatibility between
    'A[V2]' and 'A[V3]'.

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B >=1, <2"]
                    }
                ]
            })
        },
        "B": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
        },
        "C": {
            "-": wiz.definition.Definition({
                "identifier": "C",
                "requirements": ["A[V2]"]
            })
        },
        "D": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0",
                "requirements": ["A[V3]"]
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([
            Requirement("A"), Requirement("C"), Requirement("D")
        ])

    assert (
        "The dependency graph could not be resolved due to the following "
        "requirement conflicts:\n"
        "  * ::A \t[root]\n"
        "  * ::A[V2] \t[C]\n"
        "  * ::A[V3] \t[D==0.1.0]"
    ) in str(error.value)


def test_scenario_36():
    """Compute packages for the following graph.

    When parents of conflicting nodes are themselves conflicting, we attempt to
    resolve the conflicting parents before resolving the children. In this
    case, one of the conflicting children disappears while resolving parent
    conflicts.

    Root
     |
     |--(A==1.2.*): A==1.2.0
     |   |
     |   `--(B<1): B==0.1.0
     |
     `--(B): B==1.2.3
         |
         `--(A==1.4.*): A==1.4.8

    Expected: B==0.1.0, A==1.2.0

    """
    definition_mapping = {
        "A": {
            "1.4.8": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.4.8"
            }),
            "1.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.2.0",
                "requirements": ["B<1"]
            })
        },
        "B": {
            "1.2.3": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.2.3",
                "requirements": ["A==1.4.*"]
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    packages = resolver.compute_packages([
        Requirement("A==1.2.*"), Requirement("B")
    ])

    assert len(packages) == 2
    assert packages[0].identifier == "B==0.1.0"
    assert packages[1].identifier == "A==1.2.0"


def test_scenario_37():
    """Fail to compute packages for the following graph.

    Like scenario 36, when parents of conflicting nodes are themselves
    conflicting, we attempt to resolve the conflicting parents before resolving
    the children. In this case, the parent conflicts does not remove the
    children conflicts.

    Root
     |
     |--(A==1.2.*): A==1.2.0
     |   |
     |   `--(B): B==1.2.3
     |
     `--(B<1): B==0.1.0
         |
         `--(A==1.4.*): A==1.4.8

    Expected: Unable to compute due to requirement compatibility between
    'A==1.4.*' and 'A==1.2.* '.

    """
    definition_mapping = {
        "A": {
            "1.4.8": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.4.8"
            }),
            "1.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.2.0",
                "requirements": ["B"]
            })
        },
        "B": {
            "1.2.3": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.2.3",
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["A==1.4.*"]
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A==1.2.*"), Requirement("B<1")])

    assert (
        "The dependency graph could not be resolved due to the following "
        "requirement conflicts:\n"
        "  * ::A ==1.2.* \t[root]\n"
        "  * ::A ==1.4.* \t[B==0.1.0]"
    ) in str(error.value)


def test_scenario_38():
    """Fail to compute packages for the following graph.

    Like scenario 37, when parents of conflicting nodes are themselves
    conflicting, we attempt to resolve the conflicting parents before resolving
    the children. In this case, the parent conflicts cannot be solved.

    Root
     |
     |--(A==1.2.*): A==1.2.0
     |   |
     |   `--(B>1): B==1.2.3
     |
     `--(B<1): B==0.1.0
         |
         `--(A==1.4.*): A==1.4.8

    Expected: Unable to compute due to requirement compatibility between
    'A==1.4.*' and 'A==1.2.*'.

    """
    definition_mapping = {
        "A": {
            "1.4.8": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.4.8"
            }),
            "1.2.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.2.0",
                "requirements": ["B>1"]
            })
        },
        "B": {
            "1.2.3": wiz.definition.Definition({
                "identifier": "B",
                "version": "1.2.3",
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "0.1.0",
                "requirements": ["A==1.4.*"]
            })
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A==1.2.*"), Requirement("B<1")])

    assert (
        "The dependency graph could not be resolved due to the following "
        "requirement conflicts:\n"
        "  * ::A ==1.2.* \t[root]\n"
        "  * ::A ==1.4.* \t[B==0.1.0]"
    ) in str(error.value)


def test_scenario_39():
    """Compute packages for the following graph.

    Like scenario 16, combinations containing conflicts are discarded until
    a combination without conflicts can be resolved. But this time, variant
    permutation can be optimized with requirements from each variants.

    Root
     |
     |--(A): A[V1]
     |   |
     |   `--(B>=1, <2): B==1
     |
     |--(A): A[V2]
     |   |
     |   `--(B>=2, <3): B==2
     |
     |--(A): A[V3]
     |   |
     |   `--(B>=3, <4): B==3
     |
     |--(A): A[V4]
     |   |
     |   `--(B>=4, <5): B==4
     |
     |--(A): A[V5]
     |   |
     |   `--(B>=5, <6): B==5
     |
     |--(A): A[V6]
     |   |
     |   `--(B>=6, <7): B==6
     |
     |--(A): A[V7]
     |   |
     |   `--(B>=7, <8): B==7
     |
     `--(B==7): B==7

    Expected: B==1, A[V1]

    """
    definition_mapping = {
        "A": {
            "-": wiz.definition.Definition({
                "identifier": "A",
                "variants": [
                    {
                        "identifier": "V7",
                        "requirements": ["B>=7, <8"]
                    },
                    {
                        "identifier": "V6",
                        "requirements": ["B>=6, <7"]
                    },
                    {
                        "identifier": "V5",
                        "requirements": ["B>=5, <6"]
                    },
                    {
                        "identifier": "V4",
                        "requirements": ["B>=4, <5"]
                    },
                    {
                        "identifier": "V3",
                        "requirements": ["B>=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B>=2, <3"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["B>=1, <2"]
                    }
                ]
            })
        },
        "B": {
            "1": wiz.definition.Definition({
                "identifier": "B",
                "version": "1",
            }),
            "2": wiz.definition.Definition({
                "identifier": "B",
                "version": "2",
            }),
            "3": wiz.definition.Definition({
                "identifier": "B",
                "version": "3",
            }),
            "4": wiz.definition.Definition({
                "identifier": "B",
                "version": "4",
            }),
            "5": wiz.definition.Definition({
                "identifier": "B",
                "version": "5",
            }),
            "6": wiz.definition.Definition({
                "identifier": "B",
                "version": "6",
            }),
            "7": wiz.definition.Definition({
                "identifier": "B",
                "version": "7",
            }),
        },
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A"), Requirement("B==1")
    ])

    assert len(packages) == 2
    assert packages[0].identifier == "B==1"
    assert packages[1].identifier == "A[V1]"


def test_scenario_40():
    """Compute packages for the following graph.

    Like scenario 34, a conflict appear when relinking parents as two
    requirements are explicitly requesting conflicting variants: "A[V2]" and
    "A[V3]". But this time, the graph contains a version conflict which will
    be resolved by removing "A[V2]", hence making this conflict irrelevant.

    Root
     |
     |--(A[V3]): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     |--(C >=1): C==2.0.0
     |   |
     |   `--(A[V2]): A[V2]==1.0.0
     |       |
     |       `--(B >=2, <3): B==2.0.0
     |
     `--(E): E
         |
         `--(C >=1, <2): C==1.5.0

    Expected: E, C==1.5.0, B==3.0.0, A[V3]==1.0.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    }
                ]
            })
        },
        "B": {
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
        },
        "C": {
            "2.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "2.0.0",
                "requirements": ["A[V2]"]
            }),
            "1.5.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.5.0"
            }),
        },
        "E": {
            "-": wiz.definition.Definition({
                "identifier": "E",
                "requirements": ["C >=1, <2"]
            }),
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)
    packages = resolver.compute_packages([
        Requirement("A[V3]"), Requirement("C >=1"), Requirement("E"),
    ])

    assert len(packages) == 4
    assert packages[0].identifier == "E"
    assert packages[1].identifier == "C==1.5.0"
    assert packages[2].identifier == "B==3.0.0"
    assert packages[3].identifier == "A[V3]==1.0.0"


def test_scenario_41():
    """Compute packages for the following graph.

    Like scenario 34, a conflict appear when relinking parents as two
    requirements are explicitly requesting conflicting variants: "A[V2]" and
    "A[V3]". But this time, the resolver can downgrade versions of conflicting
    nodes to find a solution.

    Root
     |
     |--(A[V3]): A[V3]==1.0.0
     |   |
     |   `--(B >=3, <4): B==3.0.0
     |
     `--(C): C==1.0.0
         |
         `--(A[V2]): A[V2]==1.0.0
             |
             `--(B >=2, <3): B==2.0.0

    Expected: C==0.5.0, B==3.0.0, A[V3]==1.0.0

    """
    definition_mapping = {
        "A": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "A",
                "version": "1.0.0",
                "variants": [
                    {
                        "identifier": "V3",
                        "requirements": ["B >=3, <4"]
                    },
                    {
                        "identifier": "V2",
                        "requirements": ["B >=2, <3"]
                    }
                ]
            })
        },
        "B": {
            "3.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0"
            }),
            "2.0.0": wiz.definition.Definition({
                "identifier": "B",
                "version": "2.0.0"
            }),
        },
        "C": {
            "1.0.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.0.0",
                "requirements": ["A[V2]"]
            }),
            "0.5.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.5.0",
            })
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    packages = resolver.compute_packages([
        Requirement("A[V3]"), Requirement("C")
    ])

    assert len(packages) == 3
    assert packages[0].identifier == "C==0.5.0"
    assert packages[1].identifier == "B==3.0.0"
    assert packages[2].identifier == "A[V3]==1.0.0"


def test_scenario_42():
    """Fail to compute packages for the following graph.

    Root
     |
     |--(A): A[V1]
     |   |
     |   `--(C>1): C==1.5.0
     |
     |--(A): A[V2]
     |   |
     |   `--(C>1): C==1.5.0
     |
     |--(B): B[V1]
     |   |
     |   `--(C<1): C==0.5.0
     |
     `--(B): B[V2]
         |
         `--(C<1): C==0.5.0

    Expected: Unable to compute due to requirement compatibility between
    'C <1' and 'C >1'.

    """
    definition_mapping = {
        "A": {
            "-": wiz.definition.Definition({
                "identifier": "A",
                "variants": [
                    {
                        "identifier": "V2",
                        "requirements": ["C>1"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["C>1"]
                    },
                ]
            }),
        },
        "B": {
            "-": wiz.definition.Definition({
                "identifier": "B",
                "variants": [
                    {
                        "identifier": "V2",
                        "requirements": ["C<1"]
                    },
                    {
                        "identifier": "V1",
                        "requirements": ["C<1"]
                    },
                ]
            }),
        },
        "C": {
            "1.5.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "1.5.0"
            }),
            "0.5.0": wiz.definition.Definition({
                "identifier": "C",
                "version": "0.5.0"
            }),
        }
    }

    resolver = wiz.graph.Resolver(definition_mapping)

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        resolver.compute_packages([Requirement("A"), Requirement("B")])

    assert (
        "The dependency graph could not be resolved due to the "
        "following requirement conflicts:\n"
        "  * ::C <1 \t[B[V2]]\n"
        "  * ::C >1 \t[A[V2]]"
    ) in str(error.value)



# :coding: utf-8

import os
import shutil
import tempfile
import uuid

import pytest


@pytest.fixture()
def unique_name():
    """Return a unique name."""
    return "unique-{0}".format(uuid.uuid4())


@pytest.fixture()
def temporary_file(request):
    """Return a temporary file path."""
    file_handle, path = tempfile.mkstemp()
    os.close(file_handle)

    def cleanup():
        """Remove temporary file."""
        try:
            os.remove(path)
        except OSError:
            pass

    request.addfinalizer(cleanup)
    return path


@pytest.fixture()
def temporary_directory(request):
    """Return a temporary directory path."""
    path = tempfile.mkdtemp()

    def cleanup():
        """Remove temporary directory."""
        shutil.rmtree(path)

    request.addfinalizer(cleanup)

    return path


@pytest.fixture(autouse=True)
def logger(mocker):
    """Mock the 'wiz.logging' module and return logger."""
    import wiz.logging
    mocker.patch.object(wiz.logging, "configure")
    mocker.patch.object(wiz.logging, "root")

    mock_logger = mocker.Mock()
    mocker.patch.object(
        wiz.logging, "Logger", return_value=mock_logger
    )
    return mock_logger



# :coding: utf-8

import os
import re
import shutil

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

ROOT_PATH = os.path.dirname(os.path.realpath(__file__))
RESOURCE_PATH = os.path.join(ROOT_PATH, "resource")
SOURCE_PATH = os.path.join(ROOT_PATH, "source")
README_PATH = os.path.join(ROOT_PATH, "README.rst")

PACKAGE_NAME = "wiz"

# Read version from source.
with open(
    os.path.join(SOURCE_PATH, PACKAGE_NAME, "_version.py")
) as _version_file:
    VERSION = re.match(
        r".*__version__ = \"(.*?)\"", _version_file.read(), re.DOTALL
    ).group(1)

# Compute dependencies.
INSTALL_REQUIRES = [
    "click >= 7, < 8",
    "colorama >= 0.3.9, < 1",
    "distro >= 1.5.0, < 2",
    "packaging >= 17.1, < 18",
    "pystache >= 0.5.4, < 1",
    "sawmill >= 0.2.1, < 1",
    "six >= 1.15.0, < 2",
    "toml >= 0.10.1, < 1",
    "ujson >= 2.0.3, < 4"
]

DOC_REQUIRES = [
    "sphinx >= 1.8, < 2",
    "sphinx_rtd_theme >= 0.1.6, < 1",
    "lowdown >= 0.1.0, < 2",
    "sphinx-click >= 1.2.0"
]

TEST_REQUIRES = [
    "pytest >= 4, < 5",
    "pytest-benchmark >= 3.2.3, < 4",
    "pytest-cov >= 2, < 3",
    "pytest-mock >= 2, < 3",
    "pytest-runner >= 2.7, < 3",
    "pytest-xdist >= 1.18, < 2"
]

DEV_REQUIRES = [
    "versup >= 1.0.1",
]


class BuildExtended(build_py):
    """Custom command to build package with custom configuration and plugins."""

    # Extended options
    user_options = build_py.user_options + [
        (
            "wiz-config-file=", None,
            "Path to TOML file to embed in installed location as the default "
            "Wiz configuration."
        ),
        (
            "wiz-plugin-path=", None,
            "Path to directory containing Python Wiz plugins to embed in "
            "installed location."
        )
    ]

    # Initialize extended options.
    wiz_config_file = None
    wiz_plugin_path = None

    def run(self):
        """Run installation command."""
        build_py.run(self)

        build_path = os.path.join(self.build_lib, PACKAGE_NAME)
        config_path = os.path.join(build_path, "package_data", "config.toml")
        plugin_path = os.path.join(build_path, "package_data", "plugins")

        if self.wiz_config_file is not None:
            shutil.copy(self.wiz_config_file, config_path)

        if self.wiz_plugin_path is not None:
            for path in os.listdir(self.wiz_plugin_path):
                _path = os.path.join(self.wiz_plugin_path, path)
                shutil.copy(_path, plugin_path)


setup(
    name="wiz-env",
    version=VERSION,
    description="Environment management framework.",
    long_description=open(README_PATH).read(),
    url="https://github.com/themill/wiz",
    keywords="",
    author="The Mill",
    packages=find_packages(SOURCE_PATH),
    package_dir={
        "": "source"
    },
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    tests_require=TEST_REQUIRES,
    extras_require={
        "doc": DOC_REQUIRES,
        "test": TEST_REQUIRES,
        "dev": DOC_REQUIRES + TEST_REQUIRES + DEV_REQUIRES
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "wiz = wiz.__main__:main"
        ]
    },
    cmdclass={
        "build_py": BuildExtended,
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 "
        "(LGPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development",
    ],
)



# :coding: utf-8

from __future__ import print_function

import collections
import datetime
import os
import time
import textwrap

import click
import six
import six.moves
import ujson

import wiz.config
import wiz.definition
import wiz.exception
import wiz.filesystem
import wiz.history
import wiz.logging
import wiz.registry
import wiz.spawn
import wiz.symbol
import wiz.utility
from wiz import __version__

# Initiate logging handler to display potential warning when fetching config.
wiz.logging.configure()

#: Retrieve configuration mapping to initialize default values.
_CONFIG = wiz.config.fetch()

#: Click default context for all commands.
CONTEXT_SETTINGS = dict(
    max_content_width=_CONFIG.get("command", {}).get("max_content_width", 80),
    help_option_names=["-h", "--help"],
)


class _MainGroup(click.Group):
    """Extended click Group for Wiz command line main entry point."""

    def parse_args(self, context, arguments):
        """Update *context* from passed *arguments*.

        Record initial command from *info_name* and *args* and identify
        extra arguments after the "--" symbol.

        We cannot only rely on the 'allow_extra_args' option of the click
        Context as it fails to recognize extra arguments when placed after
        an argument with "nargs=-1".

        """
        extra_args = []

        if wiz.symbol.COMMAND_SEPARATOR in arguments:
            index = arguments.index(wiz.symbol.COMMAND_SEPARATOR)
            extra_args = arguments[index + 1:]
            arguments = arguments[:index]

        context.obj = {
            "initial_input": wiz.utility.combine_command(["wiz"] + arguments),
            "extra_arguments": extra_args
        }

        return super(_MainGroup, self).parse_args(context, arguments)


@click.group(
    context_settings=CONTEXT_SETTINGS,
    cls=_MainGroup,
    help=textwrap.dedent(
        """
        Wiz is an environment management framework which resolves an environment
        context or executes a command from one or several package requests.

        Example:

        \b
        >>> wiz use "app==2.*" my-plugin -- AppExe /path/to/script

        A command can also be executed from a resolved context via a command
        alias which is extracted from its corresponding package.

        Example:

        \b
        >>> wiz run python

        All available packages and command can be listed as follow:

        \b
        >>> wiz list package
        >>> wiz list command

        It is also possible to search a specific package or command as follow:

        \b
        >>> wiz search python

        """
    ),
)
@click.version_option(version=__version__)
@click.option(
    "-v", "--verbosity",
    help="Set the logging output verbosity.",
    type=click.Choice(wiz.logging.levels),
    default=_CONFIG.get("command", {}).get("verbosity", "info"),
    show_default=True
)
@click.option(
    "--no-local",
    help="Skip local registry.",
    is_flag=True,
    default=_CONFIG.get("command", {}).get("no_local", False),
)
@click.option(
    "--no-cwd",
    help=(
        "Do not attempt to discover definitions from current "
        "working directory within project."
    ),
    is_flag=True,
    default=_CONFIG.get("command", {}).get("no_cwd", False),
)
@click.option(
    "-rd", "--registry-depth",
    help="Maximum depth to recursively search for definitions.",
    default=_CONFIG.get("command", {}).get("depth"),
    type=int,
    metavar="NUMBER",
)
@click.option(
    "-r", "--registry",
    help="Set registry path for package definitions.",
    default=wiz.registry.get_defaults(),
    multiple=True,
    metavar="PATH",
    type=click.Path(),
    show_default=True
)
@click.option(
    "-add", "--add-registry",
    help="Prepend registry path to default registries.",
    multiple=True,
    metavar="PATH",
    type=click.Path()
)
@click.option(
    "--ignore-implicit",
    help=(
        "Do not include implicit packages (with 'auto-use' set to true) "
        "in resolved context."
    ),
    is_flag=True,
    default=_CONFIG.get("command", {}).get("ignore_implicit", False),
)
@click.option(
    "--init",
    help=(
        "Initial Environment which will be augmented by the resolved "
        "environment."
    ),
    type=lambda x: tuple(x.split("=", 1)),
    metavar="VARIABLE=VALUE",
    default=[],
    multiple=True,
)
@click.option(
    "--platform",
    metavar="PLATFORM",
    help="Override detected platform.",
    default=_CONFIG.get("command", {}).get("platform"),
)
@click.option(
    "--architecture",
    metavar="ARCHITECTURE",
    help="Override detected architecture.",
    default=_CONFIG.get("command", {}).get("architecture"),
)
@click.option(
    "--os-name",
    metavar="NAME",
    help="Override detected operating system name.",
    default=_CONFIG.get("command", {}).get("os_name"),
)
@click.option(
    "--os-version",
    metavar="VERSION",
    help="Override detected operating system version.",
    default=_CONFIG.get("command", {}).get("os_version"),
)
@click.option(
    "--record",
    help="Record resolution context process for debugging.",
    type=click.Path(exists=True)
)
@click.pass_context
def main(click_context, **kwargs):
    """Main entry point for the command line interface."""
    wiz.logging.configure()
    logger = wiz.logging.Logger(__name__ + ".main")

    if kwargs["record"] is not None:
        wiz.history.start_recording(
            command=click_context.obj["initial_input"]
        )

    # Set verbosity level.
    wiz.logging.root.handlers["stderr"].filterer.min = kwargs["verbosity"]

    # Identify system mapping.
    system_mapping = wiz.system.query(
        platform=kwargs["platform"],
        architecture=kwargs["architecture"],
        os_name=kwargs["os_name"],
        os_version=kwargs["os_version"],
    )
    logger.debug("System: {}".format(system_mapping))

    # Fetch all registries.
    registries = wiz.registry.fetch(
        kwargs["registry"] + kwargs["add_registry"],
        include_local=not kwargs["no_local"],
        include_working_directory=not kwargs["no_cwd"]
    )
    logger.debug("Registries: " + ", ".join(registries))

    # Extract initial environment.
    initial_environment = {x[0]: x[1] for x in kwargs["init"] if len(x) == 2}
    logger.debug("Initial environment: {}".format(initial_environment))

    # Update user data within click context.
    click_context.obj.update({
        "system_mapping": system_mapping,
        "registry_paths": registries,
        "registry_search_depth": kwargs["registry_depth"],
        "ignore_implicit_packages": kwargs["ignore_implicit"],
        "initial_environment": initial_environment,
        "recording_path": kwargs["record"],
    })


@main.group(
    name="list",
    help=textwrap.dedent(
        """
        Display all available commands or package definitions.

        Example:

        \b
        >>> wiz list command
        >>> wiz list package
        >>> wiz list command --all
        >>> wiz list package --all

        """
    ),
    short_help="List commands or package definitions.",
    context_settings=CONTEXT_SETTINGS
)
@click.pass_context
def wiz_list_group(click_context):
    """Group command which list available commands or package definitions."""
    # Ensure that context fail if extra arguments were passed.
    _fail_on_extra_arguments(click_context)


@wiz_list_group.command(
    name="package",
    help=textwrap.dedent(
        """
        Display all available package definitions.

        Example:

        \b
        >>> wiz list package
        >>> wiz list package --all

        """
    ),
    short_help="Display available packages.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "-a", "--all",
    help="Display all package versions, not just the latest one.",
    is_flag=True,
    default=(
        _CONFIG.get("command", {}).get("list", {}).get("package", {})
        .get("all", False)
    )
)
@click.option(
    "--no-arch",
    help="Display packages for all platforms.",
    is_flag=True,
    default=(
        _CONFIG.get("command", {}).get("list", {}).get("package", {})
        .get("no_arch", False)
    )
)
@click.pass_context
def wiz_list_package(click_context, **kwargs):
    """Command to list available package definitions."""
    package_mapping = {}

    system_mapping = (
        None if kwargs["no_arch"] else click_context.obj["system_mapping"]
    )

    for definition in wiz.definition.discover(
        click_context.obj["registry_paths"],
        system_mapping=system_mapping,
        max_depth=click_context.obj["registry_search_depth"]
    ):
        _add_to_mapping(definition, package_mapping)

    display_registries(click_context.obj["registry_paths"])

    display_package_mapping(
        package_mapping,
        click_context.obj["registry_paths"],
        all_versions=kwargs["all"],
        with_system=kwargs["no_arch"]
    )

    _export_history_if_requested(click_context)


@wiz_list_group.command(
    name="command",
    help=textwrap.dedent(
        """
        Display all available commands.

        Example:

        \b
        >>> wiz list command
        >>> wiz list command --all

        """
    ),
    short_help="Display available commands.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "-a", "--all",
    help="Display all command versions, not just the latest one.",
    is_flag=True,
    default=(
        _CONFIG.get("command", {}).get("list", {}).get("command", {})
        .get("all", False)
    )
)
@click.option(
    "--no-arch",
    help="Display commands for all platforms.",
    is_flag=True,
    default=(
        _CONFIG.get("command", {}).get("list", {}).get("command", {})
        .get("no_arch", False)
    )
)
@click.pass_context
def wiz_list_command(click_context, **kwargs):
    """Command to list available commands."""
    package_mapping = {}
    command_mapping = {}

    system_mapping = (
        None if kwargs["no_arch"] else click_context.obj["system_mapping"]
    )

    for definition in wiz.definition.discover(
        click_context.obj["registry_paths"],
        system_mapping=system_mapping,
        max_depth=click_context.obj["registry_search_depth"]
    ):
        _add_to_mapping(definition, package_mapping)

        for command in definition.command.keys():
            command_mapping.setdefault(command, [])
            command_mapping[command] = definition.qualified_identifier

    display_registries(click_context.obj["registry_paths"])

    display_command_mapping(
        command_mapping,
        package_mapping,
        click_context.obj["registry_paths"],
        all_versions=kwargs["all"],
        with_system=kwargs["no_arch"]
    )

    _export_history_if_requested(click_context)


@main.command(
    name="search",
    help=textwrap.dedent(
        """
        Search and display definitions from a list of filters.

        Example:

        \b
        >>> wiz search foo
        >>> wiz search foo --all
        >>> wiz search foo --no-arch
        >>> wiz search foo bar

        """
    ),
    short_help="Search package definitions.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "-a", "--all",
    help="Display all package versions, not just the latest one.",
    is_flag=True,
    default=_CONFIG.get("command", {}).get("search", {}).get("all", False)
)
@click.option(
    "--no-arch",
    help="Display results for all platforms.",
    is_flag=True,
    default=_CONFIG.get("command", {}).get("search", {}).get("no_arch", False)
)
@click.option(
    "-t", "--type",
    help="Set the request type.",
    type=click.Choice([
        "all",
        wiz.symbol.PACKAGE_REQUEST_TYPE,
        wiz.symbol.COMMAND_REQUEST_TYPE
    ]),
    default=_CONFIG.get("command", {}).get("search", {}).get("type", "all"),
    show_default=True
)
@click.argument(
    "filters",
    nargs=-1,
    required=True
)
@click.pass_context
def wiz_search(click_context, **kwargs):
    """Search and display definitions from request(s)."""
    logger = wiz.logging.Logger(__name__ + ".wiz_search")

    # Ensure that context fail if extra arguments were passed.
    _fail_on_extra_arguments(click_context)

    # Display registries.
    display_registries(click_context.obj["registry_paths"])

    # Fetch all definitions.
    package_mapping = {}
    command_mapping = {}

    # Keyword elements to filter.
    keywords = ["identifier", "version", "description"]

    system_mapping = (
        None if kwargs["no_arch"] else click_context.obj["system_mapping"]
    )

    _filters = kwargs["filters"]

    for definition in wiz.definition.discover(
        click_context.obj["registry_paths"],
        system_mapping=system_mapping,
        max_depth=click_context.obj["registry_search_depth"]
    ):
        values = [str(getattr(definition, keyword)) for keyword in keywords]
        values += definition.command.keys()

        if any(_filter in value for value in values for _filter in _filters):
            _add_to_mapping(definition, package_mapping)

        for command in definition.command.keys():
            if any(_filter in command for _filter in kwargs["filters"]):
                command_mapping.setdefault(command, [])
                command_mapping[command] = definition.qualified_identifier

    results_found = False

    if kwargs["type"] in ["command", "all"] and len(command_mapping) > 0:
        results_found = True

        display_command_mapping(
            command_mapping,
            package_mapping,
            click_context.obj["registry_paths"],
            all_versions=kwargs["all"],
            with_system=kwargs["no_arch"]
        )

    if kwargs["type"] in ["package", "all"] and len(package_mapping) > 0:
        results_found = True

        display_package_mapping(
            package_mapping,
            click_context.obj["registry_paths"],
            all_versions=kwargs["all"],
            with_system=kwargs["no_arch"]
        )

    if not results_found:
        logger.warning("No results found.\n")

    _export_history_if_requested(click_context)


@main.command(
    name="view",
    help=textwrap.dedent(
        """
        Display content of a package definition from definition identifier or
        command.

        Example:

        \b
        >>> wiz view foo
        >>> wiz view foo --json

        """
    ),
    short_help="View content of a package definition.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "--json",
    help="Display definition in JSON.",
    is_flag=True,
    default=_CONFIG.get("command", {}).get("view", {}).get("json_view", False),
)
@click.argument(
    "request",
    nargs=1,
    required=True
)
@click.pass_context
def wiz_view(click_context, **kwargs):
    """Display definition from identifier or command."""
    logger = wiz.logging.Logger(__name__ + ".wiz_view")

    # Ensure that context fail if extra arguments were passed.
    _fail_on_extra_arguments(click_context)

    definition_mapping = _fetch_definition_mapping_from_context(click_context)

    results_found = False

    # Display the corresponding definition if the request is a command.
    try:
        request = wiz.fetch_package_request_from_command(
            kwargs["request"], definition_mapping
        )
        definition = wiz.fetch_definition(request, definition_mapping)

    except wiz.exception.RequestNotFound as exception:
        logger.debug(
            "Impossible to query definition from command request: "
            "{}\n".format(exception)
        )

    else:
        logger.info(
            "Command found in definition: {}".format(
                definition.qualified_version_identifier
            )
        )
        results_found = True

    # Display the full definition if the request is a package.
    try:
        definition = wiz.fetch_definition(kwargs["request"], definition_mapping)

    except wiz.exception.RequestNotFound as exception:
        logger.debug(
            "Impossible to query definition from package request: "
            "{}\n".format(exception)
        )

    else:
        logger.info(
            "View definition: {}".format(
                definition.qualified_version_identifier
            )
        )

        if kwargs["json"]:
            click.echo(definition.encode())
        else:
            display_definition(definition)

        results_found = True

    # Otherwise, display a warning...
    if not results_found:
        logger.warning("No definition found.\n")

    _export_history_if_requested(click_context)


@main.command(
    name="use",
    help=textwrap.dedent(
        """
        Spawn shell with resolved context from requested packages, or run
        a command within the resolved context.

        Example:

        \b
        >>> wiz use package1>=1 package2==2.3.0 package3
        >>> wiz use package1>=1 package2==2.3.0 package3 -- app --option value
        >>> wiz use --view command

        """
    ),
    short_help="Use resolved context from package definition.",
    context_settings=dict(
        allow_extra_args=True,
        **CONTEXT_SETTINGS
    )
)
@click.option(
    "--view",
    help="Only view the resolved context without loading it.",
    is_flag=True,
    default=False
)
@click.option(
    "-mc", "--max-combinations",
    help=(
        "Maximum number of combinations which can be generated from "
        "conflicting variants during the context resolution process."
    ),
    default=_CONFIG.get("resolver", {}).get("maximum_combinations"),
    type=int,
    metavar="NUMBER",
    show_default=True
)
@click.option(
    "-ma", "--max-attempts",
    help=(
        "Maximum number of attempts to resolve the context before raising an "
        "error."
    ),
    default=_CONFIG.get("resolver", {}).get("maximum_attempts"),
    type=int,
    metavar="NUMBER",
    show_default=True
)
@click.argument(
    "requests",
    nargs=-1,
    required=True
)
@click.pass_context
def wiz_use(click_context, **kwargs):
    """Resolve and use context from command."""
    logger = wiz.logging.Logger(__name__ + ".wiz_use")

    definition_mapping = _fetch_definition_mapping_from_context(click_context)
    ignore_implicit = click_context.obj["ignore_implicit_packages"]
    environ_mapping = click_context.obj["initial_environment"]

    # Fetch extra arguments from context.
    extra_arguments = _fetch_extra_arguments(click_context)

    try:
        wiz_context = wiz.resolve_context(
            list(kwargs["requests"]), definition_mapping,
            ignore_implicit=ignore_implicit,
            environ_mapping=environ_mapping,
            maximum_combinations=kwargs["max_combinations"],
            maximum_attempts=kwargs["max_attempts"],
        )

        # Only view the resolved context without spawning a shell nor
        # running any commands.
        if kwargs["view"]:
            display_registries(wiz_context["registries"])
            display_resolved_context(wiz_context)

        # Do not execute any command if history is being recorded.
        elif not kwargs["view"] and click_context.obj["recording_path"]:
            pass

        # If no commands are indicated, spawn a shell.
        elif len(extra_arguments) == 0:
            wiz.spawn.shell(wiz_context["environ"], wiz_context["command"])

        # Otherwise, resolve the command and run it within the resolved context.
        else:
            command_elements = wiz.resolve_command(
                extra_arguments, wiz_context.get("command", {})
            )
            wiz.spawn.execute(command_elements, wiz_context["environ"])

    except wiz.exception.WizError as error:
        logger.error(str(error))
        logger.debug_traceback()

        wiz.history.record_action(
            wiz.symbol.EXCEPTION_RAISE_ACTION, error=error
        )

    except KeyboardInterrupt:
        logger.warning("Aborted.")

    _export_history_if_requested(click_context)


@main.command(
    "run",
    help=textwrap.dedent(
        """
        Run command from resolved context.

        Example:

        \b
        >>> wiz run command
        >>> wiz run command -- --option value /path/to/output

        """
    ),
    short_help="Run command from package definition.",
    context_settings=dict(
        allow_extra_args=True,
        ignore_unknown_options=True,
        **CONTEXT_SETTINGS
    )
)
@click.option(
    "--view",
    help="Only view the resolved context without loading it.",
    is_flag=True,
    default=False
)
@click.option(
    "-mc", "--max-combinations",
    help=(
        "Maximum number of combinations which can be generated from "
        "conflicting variants during the context resolution process."
    ),
    default=_CONFIG.get("resolver", {}).get("maximum_combinations"),
    type=int,
    metavar="NUMBER",
    show_default=True
)
@click.option(
    "-ma", "--max-attempts",
    help=(
        "Maximum number of attempts to resolve the context before raising an "
        "error."
    ),
    default=_CONFIG.get("resolver", {}).get("maximum_attempts"),
    type=int,
    metavar="NUMBER",
    show_default=True
)
@click.argument(
    "request",
    nargs=1,
    required=True
)
@click.pass_context
def wiz_run(click_context, **kwargs):
    """Run application from resolved context."""
    logger = wiz.logging.Logger(__name__ + ".wiz_run")

    definition_mapping = _fetch_definition_mapping_from_context(click_context)
    ignore_implicit = click_context.obj["ignore_implicit_packages"]
    environ_mapping = click_context.obj["initial_environment"]

    # Fetch extra arguments from context.
    extra_arguments = _fetch_extra_arguments(click_context)

    try:
        requirement = wiz.utility.get_requirement(kwargs["request"])
        request = wiz.fetch_package_request_from_command(
            kwargs["request"], definition_mapping
        )

        wiz_context = wiz.resolve_context(
            [request], definition_mapping,
            ignore_implicit=ignore_implicit,
            environ_mapping=environ_mapping,
            maximum_combinations=kwargs["max_combinations"],
            maximum_attempts=kwargs["max_attempts"],
        )

        # Only view the resolved context without spawning a shell nor
        # running any commands.
        if kwargs["view"]:
            display_registries(wiz_context["registries"])
            display_resolved_context(wiz_context)

        # Do not execute any command if history is being recorded.
        elif not kwargs["view"] and click_context.obj["recording_path"]:
            pass

        else:
            command_elements = wiz.resolve_command(
                [requirement.name] + extra_arguments,
                wiz_context.get("command", {})
            )
            wiz.spawn.execute(command_elements, wiz_context["environ"])

    except wiz.exception.WizError as error:
        logger.error(str(error))
        logger.debug_traceback()

        wiz.history.record_action(
            wiz.symbol.EXCEPTION_RAISE_ACTION, error=error
        )

    except KeyboardInterrupt:
        logger.warning("Aborted.")

    _export_history_if_requested(click_context)


@main.command(
    "freeze",
    help=textwrap.dedent(
        """
        Export resolved context into a package definition or a script.

        Example:

        \b
        >>> wiz freeze foo>=1 bar==2.3.0 baz -o /tmp
        >>> wiz freeze --format bash foo>=1 bar==2.3.0 baz -o /tmp
        >>> wiz freeze --format tcsh foo>=1 bar==2.3.0 baz -o /tmp

        """
    ),
    short_help="Export resolved context.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "-o", "--output",
    help="Indicate the output directory.",
    type=click.Path(),
    required=True
)
@click.option(
    "-F", "--format",
    help="Indicate the output format.",
    type=click.Choice(["wiz", "tcsh", "bash"]),
    default=_CONFIG.get("command", {}).get("freeze", {}).get("format", "wiz"),
    show_default=True
)
@click.argument(
    "requests",
    nargs=-1,
    required=True
)
@click.pass_context
def wiz_freeze(click_context, **kwargs):
    """Freeze resolved context into a package definition or a script."""
    logger = wiz.logging.Logger(__name__ + ".wiz_freeze")

    # Ensure that context fail if extra arguments were passed.
    _fail_on_extra_arguments(click_context)

    definition_mapping = _fetch_definition_mapping_from_context(click_context)
    ignore_implicit = click_context.obj["ignore_implicit_packages"]
    environ_mapping = click_context.obj["initial_environment"]

    try:
        _context = wiz.resolve_context(
            list(kwargs["requests"]), definition_mapping,
            ignore_implicit=ignore_implicit,
            environ_mapping=environ_mapping,
        )
        identifier = _query_identifier()

        if kwargs["format"] == "wiz":
            description = _query_description()
            version = _query_version()

            definition_data = {
                "identifier": identifier,
                "description": description,
                "version": str(version)
            }

            command_mapping = _context.get("command")
            if command_mapping is not None:
                definition_data["command"] = command_mapping

            environ_mapping = _context.get("environ")
            if environ_mapping is not None:
                definition_data["environ"] = environ_mapping

            wiz.export_definition(kwargs["output"], definition_data)

        elif kwargs["format"] == "bash":
            command = _query_command(_context.get("command", {}).values())
            wiz.export_script(
                kwargs["output"], "bash",
                identifier,
                environ=_context.get("environ", {}),
                command=command,
                packages=_context.get("packages")
            )

        elif kwargs["format"] == "tcsh":
            command = _query_command(_context.get("command", {}).values())
            wiz.export_script(
                kwargs["output"], "tcsh",
                identifier,
                environ=_context.get("environ", {}),
                command=command,
                packages=_context.get("packages")
            )

    except wiz.exception.WizError as error:
        logger.error(str(error))
        logger.debug_traceback()

        wiz.history.record_action(
            wiz.symbol.EXCEPTION_RAISE_ACTION, error=error
        )

    except KeyboardInterrupt:
        logger.warning("Aborted.")

    _export_history_if_requested(click_context)


@main.command(
    "install",
    help=textwrap.dedent(
        """
        Install a package definition to a registry.

        Example:

        \b
        >>> wiz install /path/to/foo.json -o /path/to/registry
        >>> wiz install /all/definitions/* -o /path/to/registry

        """
    ),
    short_help="Install a package definition to a registry.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "-o", "--output",
    help="Registry target to install the package to.",
    type=click.Path(),
    required=True
)
@click.option(
    "-f", "--overwrite",
    help="Always overwrite existing definitions.",
    is_flag=True,
    default=(
        _CONFIG.get("command", {}).get("install", {}).get("overwrite", False)
    )
)
@click.argument(
    "definitions",
    nargs=-1,
    required=True
)
@click.pass_context
def wiz_install(click_context, **kwargs):
    """Install a definition to a registry."""
    logger = wiz.logging.Logger(__name__ + ".wiz_install")

    # Ensure that context fail if extra arguments were passed.
    _fail_on_extra_arguments(click_context)

    overwrite = kwargs["overwrite"]

    callback = _CONFIG.get("callback", {}).get("install")
    if callback is None:
        logger.error("There are no available callback to install definitions.")
        return

    while True:
        try:
            callback(
                kwargs["definitions"], kwargs["output"],
                overwrite=overwrite
            )
            break

        except wiz.exception.DefinitionsExist as error:
            if not click.confirm(
                "{message}\n{definitions}\nOverwrite?".format(
                    message=str(error),
                    definitions="\n".join([
                        "- {}".format(definition)
                        for definition in error.definitions
                    ])
                )
            ):
                break

            overwrite = True

        except wiz.exception.InstallNoChanges:
            logger.warning("No changes detected in release.")
            break

        except Exception as error:
            logger.error(error)
            logger.debug_traceback()

            wiz.history.record_action(
                wiz.symbol.EXCEPTION_RAISE_ACTION, error=error
            )
            break

    _export_history_if_requested(click_context)


@main.command(
    "edit",
    help=textwrap.dedent(
        """
        Edit one or several definitions with default editor or with operation
        option(s).

        If an output is specified, the original definition(s) will not be
        mutated. Otherwise, the original definition file(s) will be updated with
        edited data.

        The edited definition(s) will be validated before export.

        Example:

        \b
        >>> wiz edit foo.json
        >>> wiz edit foo.json --output /tmp/target
        >>> wiz edit foo.json --set install-location --value /path/data
        >>> wiz edit foo.json --update environ --value '{"KEY": "VALUE"}'
        >>> wiz edit * --extend requirements --value "bar > 0.1.0"

        """
    ),
    short_help="Edit one or several definition(s).",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "--set",
    help="Set a new value to a keyword.",
    nargs=2,
    type=click.Tuple([str, lambda x: _casted_argument(x)]),
    metavar="<KEYWORD VALUE>",
    default=(None, None),
)
@click.option(
    "--update",
    help="Update mapping keyword with mapping value.",
    nargs=2,
    type=click.Tuple([str, lambda x: _casted_argument(x)]),
    metavar="<KEYWORD VALUE>",
    default=(None, None),
)
@click.option(
    "--extend",
    help="Extend list keyword with list value.",
    nargs=2,
    type=click.Tuple([str, lambda x: _casted_argument(x)]),
    metavar="<KEYWORD VALUE>",
    default=(None, None),
)
@click.option(
    "--insert",
    help="Insert value to keyword with.",
    nargs=3,
    type=click.Tuple([str, str, int]),
    metavar="<KEYWORD VALUE INDEX>",
    default=(None, None, None),
)
@click.option(
    "--remove",
    help="Remove keyword.",
    nargs=1,
    type=lambda x: x if isinstance(x, tuple) else (str(x),),
    metavar="<KEYWORD>",
    default=(None,),
)
@click.option(
    "--remove-key",
    help="Remove value from mapping keyword.",
    nargs=2,
    type=click.Tuple([str, str]),
    metavar="<KEYWORD NAME>",
    default=(None, None),
)
@click.option(
    "--remove-index",
    help="Remove index from list keyword.",
    nargs=2,
    type=click.Tuple([str, int]),
    metavar="<KEYWORD INDEX>",
    default=(None, None),
)
@click.option(
    "-f", "--overwrite",
    help="Always overwrite existing definitions.",
    is_flag=True,
    default=_CONFIG.get("command", {}).get("edit", {}).get("overwrite", False)
)
@click.option(
    "-o", "--output",
    help=(
        "Indicate an output directory for updated definition(s). "
        "By default, original definition will be modified."
    ),
    type=click.Path(),
)
@click.argument(
    "definitions",
    nargs=-1,
    required=True
)
@click.pass_context
def wiz_edit(click_context, **kwargs):
    """Edit one or several definition(s)."""
    logger = wiz.logging.Logger(__name__ + ".wiz_edit")

    # Ensure that context fail if extra arguments were passed.
    _fail_on_extra_arguments(click_context)

    # Fetch operations from arguments
    all_operations = [
        "set", "update", "extend", "insert", "remove", "remove_key",
        "remove_index"
    ]

    operations = [_id for _id in all_operations if kwargs[_id][0] is not None]

    try:
        # Fetch definitions from arguments.
        definitions = [
            wiz.load_definition(path) for path in kwargs["definitions"]
        ]

        for definition in definitions:
            label = wiz.utility.compute_label(definition)
            logger.info("Edit {}.".format(label))

            if len(operations) == 0:
                data = click.edit(definition.encode(), extension=".json")
                if data is None:
                    logger.warning("Skip edition for {}.".format(label))
                    continue

                definition = wiz.definition.Definition(
                    ujson.loads(data), path=definition.path
                )

            else:
                for name in operations:
                    args = kwargs[name]
                    definition = getattr(definition, name)(*args)

            path = definition.path

            if kwargs["output"] is not None:
                name = os.path.basename(definition.path)
                path = os.path.join(kwargs["output"], name)

            overwrite = kwargs["overwrite"]

            while True:
                try:
                    wiz.filesystem.export(
                        path, definition.encode(), overwrite=overwrite
                    )
                    logger.info("Saved {} in {}.".format(label, path))
                    break

                except wiz.exception.FileExists:
                    if not click.confirm("Overwrite {}?".format(label)):
                        logger.warning("Skip edition for {}.".format(label))
                        break

                    overwrite = True

    except Exception as error:
        logger.error(str(error))
        logger.debug_traceback()

        wiz.history.record_action(
            wiz.symbol.EXCEPTION_RAISE_ACTION, error=error
        )

    _export_history_if_requested(click_context)


@main.command(
    "analyze",
    help=textwrap.dedent(
        """
        Analyze reachable definitions and display corresponding errors and
        warnings. A filter can be set to target specific definitions.

        Example:

        \b
        >>> wiz analyze
        >>> wiz analyze "foo" "bar"
        >>> wiz analyze --verbose
        >>> wiz -r /path/to/registry analyze
        >>> wiz -add /path/to/additional/registry analyze

        """
    ),
    short_help="Analyze reachable definitions.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "--no-arch",
    help="Analyze definition for all platforms.",
    is_flag=True,
    default=_CONFIG.get("command", {}).get("analyze", {}).get("no_arch", False)
)
@click.option(
    "--verbose",
    help="Increase verbosity of analysis.",
    is_flag=True,
    default=_CONFIG.get("command", {}).get("analyze", {}).get("verbose", False)
)
@click.argument(
    "filters",
    nargs=-1,
)
@click.pass_context
def wiz_analyze(click_context, **kwargs):
    """Display warning and error for each registry."""
    logger = wiz.logging.Logger(__name__ + ".wiz_analyze")
    if click_context.obj["recording_path"] is not None:
        logger.error("Impossible to record history during analysis.")
        return

    definition_mapping = _fetch_definition_mapping_from_context(click_context)
    system_mapping = (
        None if kwargs["no_arch"] else click_context.obj["system_mapping"]
    )

    latest_registry = None

    for definition in wiz.definition.discover(
        click_context.obj["registry_paths"],
        system_mapping=system_mapping,
        max_depth=click_context.obj["registry_search_depth"]
    ):
        if latest_registry != definition.registry_path:
            info = "\nRegistry: {}\n".format(definition.registry_path)
            print(wiz.utility.colored_text(info, color="cyan"))
            latest_registry = definition.registry_path

        identifier = definition.qualified_version_identifier

        # Skip definition if not matching filters.
        if any(
            _filter.lower() not in identifier.lower()
            for _filter in kwargs["filters"]
        ):
            continue

        print("  - {}".format(identifier), end="")
        if kwargs["no_arch"]:
            system_label = wiz.utility.compute_system_label(definition)
            print(" [{}]".format(system_label), end="")

        display_definition_analysis(
            definition,
            definition_mapping=definition_mapping,
            verbose=kwargs["verbose"],
        )

    if latest_registry is None:
        print(wiz.utility.colored_text("No definitions found.", color="red"))

    print()


def display_definition_analysis(
    definition, definition_mapping=None, verbose=False
):
    """Analyze *definition* and display results.

    Example::

        >>> display_definition_analysis(definition)

          - foo==10.0 ✘
        critical: Failed to resolve graph at combination #10:

        The dependency graph could not be resolved due to the following
        requirement conflicts:
          * foo ==1.5 	    [bar[1.5]==2.0.3]
          * foo >=1, <1.5 	[bim==3.0.0]

        >>> display_definition_analysis(definition, verbose=True)

          - foo==10.0 ✘
        critical: Failed to resolve graph at combination #10:

        The dependency graph could not be resolved due to the following
        requirement conflicts:
          * foo ==1.5 	    [bar[1.5]==2.0.3]
          * foo >=1, <1.5 	[bim==3.0.0]

        * TIME DURATION: 0.276069879532
        * CREATE_GRAPH: 1
        * UPDATE_GRAPH: 6
        * CREATE_NODE: 37
        * CREATE_LINK: 98
        * CREATE_DISTANCE_MAPPING: 15
        * IDENTIFY_VARIANT_CONFLICTS: 3
        * EXTRACT_GRAPH_COMBINATION: 10
        * REMOVE_NODE: 25
        * IDENTIFY_VERSION_CONFLICTS: 1
        * RESOLUTION_ERROR: 10
        * REPLACE_NODES: 3

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :param definition_mapping: Mapping regrouping all available definitions. It
        could be fetched with :func:`wiz.fetch_definition_mapping`. If no
        definition mapping is provided, a default one will be fetched from
        :func:`default registries <wiz.registry.get_defaults>`.

    :param verbose: Indicate whether time duration and history information
        should be added to analysis.

    """
    if definition_mapping is None:
        definition_mapping = wiz.fetch_definition_mapping(
            wiz.registry.get_defaults()
        )

    if verbose:
        wiz.history.start_recording(minimal_actions=True)

    time_start = time.time()

    mapping = wiz.validate_definition(
        definition, definition_mapping=definition_mapping
    )

    time_duration = time.time() - time_start

    errors = mapping.get("errors", [])
    warnings = mapping.get("warnings", [])

    if len(errors) > 0 or len(warnings) > 0:
        print(wiz.utility.colored_text(" ✘", color="red"))

        if len(errors) > 0:
            _errors = "\n".join(errors)
            print(wiz.utility.colored_text(_errors, color="red"))

        if len(warnings) > 0:
            _warnings = "\n".join(warnings)
            print(wiz.utility.colored_text(_warnings, color="yellow"))

    else:
        print(wiz.utility.colored_text(" ✔", color="green"))

    if verbose:
        _mapping = collections.OrderedDict({"TIME DURATION": time_duration})
        history = wiz.history.get()

        for action in history.get("actions", []):
            identifier = action.get("identifier")
            _mapping.setdefault(identifier, 0)
            _mapping[identifier] += 1

        message = "\n".join([
            "    * {}: {}".format(key, value)
            for key, value in _mapping.items()
        ])

        print(wiz.utility.colored_text(message + "\n", color="green"))


def display_registries(paths):
    """Display *paths* for each registry.

    Example::

        >>> display_registries(paths)

        [0] /path/to/registry-1
        [1] /path/to/registry-2

    :param paths: List of registry paths.

    """
    columns = _create_columns(["Registries"])

    for index, path in enumerate(paths):
        _create_row("[{}] {}".format(index, path), columns[0])

    if len(paths) == 0:
        message = "No registries to display."
        _create_row(message, columns[0], resize=False)

    _display_table(columns)


def display_definition(definition):
    """Display *definition*.

    Example::

        >>> display_definition(definition)

        path: /path/to/foo.json
        registry: /path/to/registry
        identifier: Foo
        version: 0.1.0
        description: Description of Foo
        install-location: /path/to/foo
        system:
            os: el >= 7, < 8
            arch: x86_64
        command:
            foo: FooExe
        environ:
            PATH: ${INSTALL_LOCATION}/bin:${PATH}
            LD_LIBRARY_PATH: ${INSTALL_LOCATION}/lib:${LD_LIBRARY_PATH}

    :param definition: Instance of :class:`wiz.definition.Definition`.

    """

    def _display(item, level=0):
        """Display *item*"""
        indent = " " * level

        if isinstance(item, collections.OrderedDict) or isinstance(item, dict):
            for key, value in item.items():
                if (
                    isinstance(value, six.string_types) or
                    isinstance(value, int)
                ):
                    click.echo("{}{}: {}".format(indent, key, value))
                else:
                    click.echo("{}{}:".format(indent, key))
                    _display(value, level=level + 4)

        elif isinstance(item, list) or isinstance(item, tuple):
            for _item in item:
                if isinstance(_item, collections.OrderedDict):
                    click.echo("{}identifier: {}".format(
                        indent, _item.pop("identifier")
                    ))
                    _display(_item, level=level + 4)
                else:
                    _display(_item, level=level)

        else:
            click.echo("{}{}".format(indent, item))

    click.echo("path: {}".format(definition.path))
    click.echo("registry: {}".format(definition.registry_path))
    _display(definition.ordered_data())


def display_command_mapping(
    command_mapping, package_mapping, registries, all_versions=False,
    with_system=False
):
    """Display command mapping.

    Example::

        >>> display_command_mapping(
        ...     command_mapping, package_mapping, registries, all_versions=True
        ... )

        Command   Version   System                Registry   Description
        -------   -------   -------------------   --------   ------------------
        fooExe    1.0.0     noarch                0          Description of Foo
        fooExe    0.1.0     noarch                0          Description of Foo
        python    3.6.6     linux : el >= 6, <7   1          Python interpreter
        python    2.7.4     linux : el >= 6, <7   1          Python interpreter

    :param command_mapping: Mapping which associates all available commands with
        a package definition. It should be in the form of::

            {
                "fooExe": "foo",
                ...
            }

    :param package_mapping: Mapping which associates each package definition
        with an identifier, a version and a system label. It should be in the
        form of::

            {
                "foo": {
                    "1.1.0": {
                        "linux : el >=6, <7": <Definition(identifier="foo")>,
                        "linux : el >=7, <8": <Definition(identifier="foo")>,
                    "0.1.0": {
                        "linux : el >=6, <7": <Definition(identifier="foo")>,
                    ...
                },
                ...
            }

    :param registries: List of registry paths from which definitions were
        fetched.

    :param all_versions: Indicate whether all package definition versions should
        be displayed. Default is False, which means that only the latest version
        will be displayed.

    :param with_system: Indicate whether the package system should be displayed.
        Default is False.

    """
    headers = ["Command", "Version", "Registry", "Description"]
    if with_system:
        headers.insert(2, "System")

    columns = _create_columns(headers)

    success = False

    for command, identifier in sorted(command_mapping.items()):
        versions = sorted(package_mapping[identifier].keys(), reverse=True)

        # Filter latest version if requested.
        if not all_versions:
            versions = [versions[0]]

        for version in versions:
            for system_label in sorted(
                package_mapping[identifier][version].keys()
            ):
                definition = package_mapping[identifier][version][system_label]

                identifiers = [
                    "{} [{}]".format(command, variant.identifier)
                    for variant in definition.variants
                ] or [command]

                for _identifier in identifiers:
                    rows = [
                        _identifier,
                        definition.version,
                        registries.index(definition.registry_path),
                        definition.description
                    ]

                    if with_system:
                        rows.insert(2, system_label)

                    for index, row in enumerate(rows):
                        _create_row(row, columns[index])

        success = True

    if not success:
        message = "No commands to display."
        _create_row(message, columns[0], resize=False)

    _display_table(columns)


def display_package_mapping(
    package_mapping, registries, all_versions=False, with_system=False
):
    """Display package mapping

    Example::

        >>> display_command_mapping(
        ...     package_mapping, registries, all_versions=True
        ... )

        Package    Version   System                Registry   Description
        --------   -------   -------------------   --------   ------------------
        foo        1.0.0     noarch                0          Description of Foo
        foo        0.1.0     noarch                0          Description of Foo
        bar [V1]   0.1.0     linux                 0          Description of Bar
        bar [V2]   0.1.0     linux                 0          Description of Bar
        bar [V3]   0.1.0     linux                 0          Description of Bar
        python     3.6.6     linux : el >= 6, <7   0          Python interpreter
        python     2.7.4     linux : el >= 6, <7   0          Python interpreter

    :param package_mapping: Mapping which associates each package definition
        with an identifier, a version and a system label. It should be in the
        form of::

            {
                "foo": {
                    "1.1.0": {
                        "linux : el >=6, <7": <Definition(identifier="foo")>,
                        "linux : el >=7, <8": <Definition(identifier="foo")>,
                    "0.1.0": {
                        "linux : el >=6, <7": <Definition(identifier="foo")>,
                    ...
                },
                ...
            }

    :param registries: List of registry paths from which definitions were
        fetched.

    :param all_versions: Indicate whether all package definition versions should
        be displayed. Default is False, which means that only the latest version
        will be displayed.

    :param with_system: Indicate whether the package system should be displayed.
        Default is False.

    """
    headers = ["Package", "Version", "Registry", "Description"]
    if with_system:
        headers.insert(2, "System")

    columns = _create_columns(headers)

    success = False

    for identifier in sorted(package_mapping.keys()):
        versions = sorted(package_mapping[identifier].keys(), reverse=True)

        # Filter latest version if requested.
        if not all_versions:
            versions = [versions[0]]

        for version in versions:
            for system_label in sorted(
                package_mapping[identifier][version].keys()
            ):
                definition = package_mapping[identifier][version][system_label]

                identifiers = [
                    "{} [{}]".format(identifier, variant.identifier)
                    for variant in definition.variants
                ] or [identifier]

                for _identifier in identifiers:
                    rows = [
                        _identifier,
                        definition.version,
                        registries.index(definition.registry_path),
                        definition.description
                    ]

                    if with_system:
                        rows.insert(2, system_label)

                    for index, row in enumerate(rows):
                        _create_row(row, columns[index])

        success = True

    if not success:
        message = "No packages to display."
        _create_row(message, columns[0], resize=False)

    _display_table(columns)


def display_resolved_context(context):
    """Display resolved *context* mapping.

    Example::

        >>> display_resolved_context(context)

        Package   Version   Registry   Description
        -------   -------   --------   -------------------
        foo       0.1.0     0          Description of Foo.
        bar       1.1.0     0          Description of Bar.

        Command   Value
        -------   ------
        foo       fooExe


        Environment Variable   Environment Value
        --------------------   --------------------
        HOME                   /usr/people/john-doe
        LD_LIBRARY_PATH        /path/to/foo/lib
                               /path/to/bar/lib
        PATH                   /path/to/foo/bin
                               /path/to/bar/bin
                               /usr/local/sbin
                               /usr/local/bin
                               /usr/sbin
                               /usr/bin
                               /sbin
                               /bin
        USER                   john-doe

    :param context: Context mapping as resolved by :func:`wiz.resolve_context`.

    """
    _display_packages_from_context(context)
    _display_command_from_context(context)
    _display_environ_from_context(context)


def _display_packages_from_context(context):
    """Display packages contained in *context* mapping.

    :param context: Context mapping as resolved by :func:`wiz.resolve_context`.

    """
    columns = _create_columns(["Package", "Version", "Registry", "Description"])
    registries = context.get("registries", [])

    success = False

    for package in context.get("packages", []):
        _identifier = package.definition.qualified_identifier
        if package.variant_identifier is not None:
            _identifier += " [{}]".format(package.variant_identifier)

        rows = [
            _identifier,
            package.version,
            registries.index(package.definition.registry_path),
            package.description
        ]

        for index, row in enumerate(rows):
            _create_row(row, columns[index])

        success = True

    if not success:
        message = "No packages to display."
        _create_row(message, columns[0], resize=False)

    _display_table(columns)


def _display_command_from_context(context):
    """Display commands contained in *context* mapping.

    :param context: Context mapping as resolved by :func:`wiz.resolve_context`.

    """
    columns = _create_columns(["Command", "Value"])

    command_mapping = context.get("command", {})

    success = False

    for command, value in sorted(command_mapping.items()):
        _create_row(command, columns[0])
        _create_row(value, columns[1])

        success = True

    if not success:
        message = "No commands to display."
        _create_row(message, columns[0], resize=False)

    _display_table(columns)


def _display_environ_from_context(context):
    """Display environment variables contained in *context* mapping.

    :param context: Context mapping as resolved by :func:`wiz.resolve_context`.

    """
    columns = _create_columns(["Environment Variable", "Environment Value"])

    environ_mapping = context.get("environ", {})

    def _compute_value(_variable, value):
        """Compute value to display."""
        if _variable == "DISPLAY":
            return [value]
        if _variable == "WIZ_CONTEXT":
            return [str(value[:50]) + "..."]
        return str(value).split(os.pathsep)

    success = False

    for variable in sorted(environ_mapping.keys()):
        for key, _value in six.moves.zip_longest(
            [variable], _compute_value(variable, environ_mapping[variable])
        ):
            _create_row(key or "", columns[0])
            _create_row(_value, columns[1])

        success = True

    if not success:
        message = "No environment variables to display."
        _create_row(message, columns[0], resize=False)

    _display_table(columns)


def _casted_argument(argument):
    """Return *argument* casted into a proper type from JSON decoder."""
    # Ensure that boolean value are in JSON format.
    argument = argument.replace("True", "true").replace("False", "false")

    try:
        return ujson.loads(argument)
    except ValueError:
        return argument


def _query_identifier():
    """Query an identifier for a resolved context."""
    logger = wiz.logging.Logger(__name__ + "._query_identifier")

    while True:
        value = click.prompt("Please enter a definition identifier")
        identifier = wiz.filesystem.sanitize_value(value.strip())
        if len(identifier) > 2:
            return identifier

        logger.warning(
            "'{}' is an incorrect identifier (It must be at least 3 "
            "characters long.)".format(identifier)
        )


def _query_description():
    """Query an description for a resolved context."""
    logger = wiz.logging.Logger(__name__ + "._query_description")

    while True:
        value = click.prompt("Please enter a description")
        description = value.strip()

        if len(description) > 5:
            return description

        logger.warning(
            "'{}' is an incorrect description (It must be at least 5 "
            "characters long.)".format(description)
        )


def _query_version(default="0.1.0"):
    """Query a version for a resolved context."""
    logger = wiz.logging.Logger(__name__ + "._query_version")

    while True:
        content = click.prompt("Please indicate a version", default=default)

        try:
            version = wiz.utility.get_version(content)
        except wiz.exception.VersionError as error:
            logger.warning(error)
            continue
        else:
            return version


def _query_command(aliases=None):
    """Query the commands to run within the exported wrapper."""
    if aliases is not None and len(aliases) > 0:
        click.echo("Available aliases:")
        for _command in sorted(aliases):
            click.echo("- {}".format(_command))

    return click.prompt(
        "Please indicate a command if necessary", default=None
    )


def _create_columns(titles):
    """Create columns from *titles*."""
    return [
        {"size": len(title), "rows": [], "title": title} for title in titles
    ]


def _create_row(element, column, resize=True):
    """Add row with *element* in *column*."""
    _element = str(element)
    column["rows"].append(_element)

    if resize:
        column["size"] = max(len(_element), column["size"])


def _display_table(columns):
    """Display table of *rows*."""
    spaces = []
    for column in columns:
        space = column["size"] - len(column["title"])
        spaces.append(space)

    # Print titles.
    click.echo(
        "\n" + "   ".join([
            columns[i]["title"] + " " * spaces[i]
            for i in range(len(columns))
        ])
    )

    # Print underlines.
    click.echo(
        "   ".join([
            "-" * (len(columns[i]["title"]) + spaces[i])
            for i in range(len(columns))
        ])
    )

    # Print elements.
    for row in six.moves.zip_longest(*[column["rows"] for column in columns]):
        click.echo(
            "   ".join([
                row[i] + " " * (columns[i]["size"] - len(row[i]))
                for i in range(len(row))
                if row[i] is not None
            ])
        )

    # Print final blank line.
    click.echo()


def _add_to_mapping(definition, mapping):
    """Mutate package *mapping* to add *definition*

    The mutated mapping should be in the form of::

        {
            "foo": {
                "1.1.0": {
                    "linux : el >=6, <7": <Definition(identifier="foo")>,
                    "linux : el >=7, <8": <Definition(identifier="foo")>,
                "0.1.0": {
                    "linux : el >=6, <7": <Definition(identifier="foo")>,
                ...
            },
            "bar": {
                "1.1.0": {
                    "noarch": <Definition(identifier="bar")>
                }
                ...
            },
            ...
        }

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :param mapping: Mapping to mutate.

    """
    identifier = definition.qualified_identifier
    version = str(definition.version)
    system_label = wiz.utility.compute_system_label(definition)

    mapping.setdefault(identifier, {})
    mapping[identifier].setdefault(version, {})
    mapping[identifier][version].setdefault(system_label, {})
    mapping[identifier][version][system_label] = definition


def _fail_on_extra_arguments(click_context):
    """Raise an error if extra arguments are found in command."""
    arguments = _fetch_extra_arguments(click_context)
    if len(arguments):
        click_context.fail(
            "Got unexpected extra argument(s) ({})".format(" ".join(arguments))
        )


def _fetch_extra_arguments(click_context):
    """Return extra arguments from context.

    If extra arguments have not been recorded after the "--" symbol, we check
    if leftover arguments remain in context.

    """
    return click_context.obj["extra_arguments"] or click_context.args


def _fetch_definition_mapping_from_context(click_context):
    """Return definition mapping from elements stored in *click_context*."""
    return wiz.fetch_definition_mapping(
        click_context.obj["registry_paths"],
        system_mapping=click_context.obj["system_mapping"],
        max_depth=click_context.obj["registry_search_depth"]
    )


def _export_history_if_requested(click_context):
    """Return definition mapping from elements stored in *click_context*."""
    logger = wiz.logging.Logger(__name__ + "._export_history_if_requested")

    if click_context.obj["recording_path"] is None:
        return

    history = wiz.history.get(serialized=True)
    path = os.path.join(
        os.path.abspath(click_context.obj["recording_path"]),
        "wiz-{}.dump".format(datetime.datetime.now().isoformat())
    )
    wiz.filesystem.export(path, history, compressed=True)
    logger.info("History recorded and exported in '{}'".format(path))



# :coding: utf-8

import getpass
import os
import socket

#: Unique identifier of the plugin.
IDENTIFIER = "environ"


def register(config):
    """Register initial environment variables.

    The initial environment mapping contains basic variables from the external
    environment that can be used by the resolved environment, such as
    the *USER*, the *HOSTNAME* or the *HOME* variables.

    The *PATH* variable is initialized with default values to have access to the
    basic UNIX commands.

    """
    environ = {
        "USER": getpass.getuser(),
        "HOME": os.path.expanduser("~"),
        "HOSTNAME": socket.gethostname(),
        "PATH": os.pathsep.join([
            "/usr/local/sbin",
            "/usr/local/bin",
            "/usr/sbin",
            "/usr/bin",
            "/sbin",
            "/bin",
        ])
    }

    config.setdefault("environ", {})
    config["environ"].setdefault("initial", {})
    config["environ"]["initial"].update(environ)
